# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for creating GHZ circuits for benchmarking purposes."""

from typing import Optional

import cirq

from mitiq import QPROGRAM
from mitiq.interface import convert_from_mitiq


def generate_ghz_circuit(
    n_qubits: int,
    return_type: Optional[str] = None,
) -> QPROGRAM:
    """Returns a GHZ circuit ie a circuit that prepares an ``n_qubits``
    GHZ state.

    Args:
        n_qubits: The number of qubits in the circuit.
        return_type: String which specifies the type of the returned
            circuits. See the keys of ``mitiq.SUPPORTED_PROGRAM_TYPES``
            for options. If ``None``, the returned circuits have type
            ``cirq.Circuit``.

    Returns:
        A GHZ circuit acting on ``n_qubits`` qubits.
    """
    if n_qubits <= 0:
        raise ValueError(
            "Cannot prepare a GHZ circuit with {} qubits", n_qubits
        )

    qubits = cirq.LineQubit.range(n_qubits)
    circuit = cirq.Circuit()
    circuit.append(cirq.H(qubits[0]))
    for i in range(0, n_qubits - 1):
        circuit.append(cirq.CNOT(qubits[i], qubits[i + 1]))

    return_type = "cirq" if not return_type else return_type

    return convert_from_mitiq(circuit, return_type)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from collections import Counter
from numbers import Number
from typing import Any, Dict, List, Optional, Sequence, Set, Union, cast
from typing import Counter as TCounter

import cirq
import numpy as np
import numpy.typing as npt

from mitiq import QPROGRAM, MeasurementResult
from mitiq.interface import atomic_converter
from mitiq.utils import _cirq_pauli_to_string


class PauliString:
    """A `PauliString` is a (tensor) product of single-qubit Pauli gates I, X,
    Y, and Z, with a leading (real or complex) coefficient. `PauliString`s can
    be measured in any `mitiq.QPROGRAM`.
    """

    _string_to_gate_map = {"I": cirq.I, "X": cirq.X, "Y": cirq.Y, "Z": cirq.Z}

    def __init__(
        self,
        spec: str = "",
        coeff: complex = 1.0,
        support: Optional[Sequence[int]] = None,
    ) -> None:
        """Initialize a PauliString.

        Args:
            spec: String specifier of the PauliString. Should only contain
                characters 'I', 'X', 'Y', and 'Z'.
            coeff: Coefficient of the PauliString.
            support: Qubits the ``spec`` acts on, if provided.

        Examples:
            >>> PauliString(spec="IXY")  # X(1)*Y(2)
            >>> PauliString(spec="ZZ", coeff=-0.5)  # -0.5*Z(0)*Z(1)
            >>> PauliString(spec="XZ", support=(10, 17))  # X(10)*Z(17)
        """
        if not set(spec).issubset(set(self._string_to_gate_map.keys())):
            raise ValueError(
                f"One or more invalid characters in spec {spec}. Valid "
                f"characters are 'I', 'X', 'Y', and 'Z', and the spec should "
                f"not contain any spaces."
            )
        if support is not None:
            if len(support) != len(spec):
                raise ValueError(
                    f"The spec has {len(spec)} Pauli's but the support has "
                    f"{len(support)} qubits. These numbers must be equal."
                )
        else:
            support = range(len(spec))

        self._pauli: cirq.PauliString[Any] = cirq.PauliString(
            coeff,
            (
                self._string_to_gate_map[s].on(cirq.LineQubit(i))
                for (i, s) in zip(support, spec)
            ),
        )

    @staticmethod
    def from_cirq_pauli_string(
        cirq_pauli_string: cirq.PauliString[Any],
    ) -> "PauliString":
        return PauliString(
            spec=_cirq_pauli_to_string(cirq_pauli_string),
            coeff=cirq_pauli_string.coefficient,  # type: ignore
            support=sorted(q.x for q in cirq_pauli_string.qubits),
        )

    @property
    def coeff(self) -> complex:
        return cast(complex, self._pauli.coefficient)

    def matrix(
        self,
        qubit_indices_to_include: Optional[List[int]] = None,
    ) -> npt.NDArray[np.complex64]:
        """Returns the (potentially very large) matrix of the PauliString."""
        qubits = (
            [cirq.LineQubit(x) for x in qubit_indices_to_include]
            if qubit_indices_to_include
            else self._pauli.qubits
        )
        return self._pauli.matrix(qubits=qubits)

    def _basis_rotations(self) -> List[cirq.Operation]:
        """Returns the basis rotations needed to measure the PauliString."""
        return [
            op
            for op in self._pauli.to_z_basis_ops()
            if op.gate != cirq.SingleQubitCliffordGate.I
        ]

    def _qubits_to_measure(self) -> Set[cirq.Qid]:
        return set(self._pauli.qubits)

    def measure_in(self, circuit: QPROGRAM) -> QPROGRAM:
        return PauliStringCollection(self).measure_in(circuit)

    def can_be_measured_with(self, other: "PauliString") -> bool:
        """Returns True if the expectation value of the PauliString can be
        simultaneously estimated with `other` via single-qubit measurements.

        Args:
            other: The PauliString to check simultaneous measurement with.
        """
        overlap = set(self._pauli.qubits).intersection(
            set(other._pauli.qubits)
        )
        for qubit in overlap:
            if cirq.I in (self._pauli.get(qubit), other._pauli.get(qubit)):
                continue
            if self._pauli.get(qubit) != other._pauli.get(qubit):
                return False
        return True

    def with_coeff(self, coeff: complex) -> "PauliString":
        return PauliString(
            spec=self.spec, coeff=coeff, support=sorted(self.support())
        )

    @property
    def spec(self) -> str:
        """Returns a string representation of the Pauli gates in
        the PauliString."""
        return _cirq_pauli_to_string(self._pauli)

    def support(self) -> Set[int]:
        return {q.x for q in self._pauli.qubits}

    def weight(self) -> int:
        """Returns the weight of the PauliString, i.e., the number of
        non-identity terms in the PauliString.
        """
        return sum(gate != cirq.I for gate in self._pauli.values())

    def _expectation_from_measurements(
        self, measurements: MeasurementResult
    ) -> float:
        return PauliStringCollection(self)._expectation_from_measurements(
            measurements
        )

    def __mul__(self, other: Union["PauliString", Number]) -> "PauliString":
        if isinstance(other, PauliString):
            return PauliString.from_cirq_pauli_string(
                self._pauli * other._pauli
            )
        elif isinstance(other, Number):
            return PauliString.from_cirq_pauli_string(self._pauli * other)
        return NotImplemented

    def __rmul__(self, other: Number) -> "PauliString":
        if isinstance(other, Number):
            return self.__mul__(other)
        return NotImplemented

    def __eq__(self, other: Any) -> bool:
        return self._pauli == other._pauli

    def __hash__(self) -> int:
        return self._pauli.__hash__()

    def __str__(self) -> str:
        return str(self._pauli)

    def __repr__(self) -> str:
        return repr(self._pauli)


class PauliStringCollection:
    """A collection of PauliStrings that qubit-wise commute and so can be
    measured with a single circuit.
    """

    def __init__(
        self, *paulis: PauliString, check_precondition: bool = True
    ) -> None:
        """Initializes a `PauliStringCollection`.

        Args:
            paulis: PauliStrings to add to the collection.
            check_precondition: If True, raises an error if some of the
                `PauliString`s do not qubit-wise commute.

        Example:
            >>> pcol = PauliStringCollection(
            >>>     PauliString(spec="X"),
            >>>     PauliString(spec="IZ", coeff=-2.2)
            >>> )
            >>> print(pcol)  # X(0) + (-2.2+0j)*Z(1)
            >>> print(pcol.support())  # {0, 1}
            >>>
            >>> # XZ qubit-wise commutes with X(0) and Z(1), so can be added.
            >>> print(pcol.can_add(PauliString(spec="XZ")))  # True.
            >>> pcol.add(PauliString(spec="XZ"))
            >>> print(pcol)  # X(0) + (-2.2+0j)*Z(1) + X(0)*Z(1)
            >>>
            >>> # Z(0) doesn't qubit-wise commute with X(0), so can't be added.
            >>> print(pcol.can_add(PauliString(spec="Z")))  # False.
        """
        self._paulis_by_weight: Dict[int, TCounter[PauliString]] = dict()
        self.add(*paulis, check_precondition=check_precondition)

    def can_add(self, pauli: PauliString) -> bool:
        return all(pauli.can_be_measured_with(p) for p in self.elements)

    def add(
        self, *paulis: PauliString, check_precondition: bool = True
    ) -> None:
        for pauli in paulis:
            if check_precondition and not self.can_add(pauli):
                raise ValueError(
                    f"Cannot add PauliString {pauli} to PauliStringCollection."
                )
            weight = pauli.weight()
            if self._paulis_by_weight.get(weight) is None:
                self._paulis_by_weight[weight] = Counter({pauli})
            else:
                self._paulis_by_weight[weight].update({pauli})

    @property
    def elements(self) -> List[PauliString]:
        return [
            pauli
            for paulis in self._paulis_by_weight.values()
            for pauli in paulis.elements()
        ]

    @property
    def elements_by_weight(self) -> Dict[int, TCounter[PauliString]]:
        return self._paulis_by_weight

    def support(self) -> Set[int]:
        return {cast(cirq.LineQubit, q).x for q in self._qubits_to_measure()}

    def max_weight(self) -> int:
        return max(self._paulis_by_weight.keys(), default=0)

    def min_weight(self) -> int:
        return min(self._paulis_by_weight.keys(), default=0)

    def _qubits_to_measure(self) -> Set[cirq.Qid]:
        qubits: Set[cirq.Qid] = set()
        for pauli in self.elements:
            qubits.update(pauli._pauli.qubits)
        return qubits

    def measure_in(self, circuit: QPROGRAM) -> QPROGRAM:
        return self._measure_in(circuit, self)

    @staticmethod
    @atomic_converter
    def _measure_in(
        circuit: cirq.Circuit, paulis: "PauliStringCollection"
    ) -> cirq.Circuit:
        # Transform circuit to canonical qubit layout.
        qubit_map = dict(
            zip(
                sorted(circuit.all_qubits()),
                cirq.LineQubit.range(len(circuit.all_qubits())),
            )
        )
        circuit = circuit.transform_qubits(lambda q: qubit_map[q])

        if not paulis._qubits_to_measure().issubset(set(circuit.all_qubits())):
            raise ValueError(
                f"Qubit mismatch. The PauliString(s) act on qubits "
                f"{paulis.support()} but the circuit has qubit indices "
                f"{sorted([q for q in circuit.all_qubits()])}."
            )

        basis_rotations = set()
        support = set()
        for pauli in paulis.elements:
            basis_rotations.update(pauli._basis_rotations())
            support.update(pauli._qubits_to_measure())
        measured = circuit + basis_rotations + cirq.measure(*sorted(support))

        # Transform circuit back to original qubits.
        reverse_qubit_map = dict(zip(qubit_map.values(), qubit_map.keys()))
        return measured.transform_qubits(lambda q: reverse_qubit_map[q])

    def _expectation_from_measurements(
        self, measurements: MeasurementResult
    ) -> float:
        total = 0.0
        for pauli in self.elements:
            bitstrings = measurements.filter_qubits(sorted(pauli.support()))
            value = (
                np.average([(-1) ** np.sum(bits) for bits in bitstrings])
                if len(bitstrings) > 0
                else 1.0
            )
            total += pauli.coeff * value
        return total

    def __eq__(self, other: Any) -> bool:
        return self._paulis_by_weight == other._paulis_by_weight

    def __len__(self) -> int:
        return len(self.elements)

    def __str__(self) -> str:
        return " + ".join(map(str, self.elements))



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""High-level digital dynamical decoupling (DDD) tools."""

from functools import partial, wraps
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np

from mitiq import QPROGRAM, Executor, Observable, QuantumResult
from mitiq.ddd.insertion import insert_ddd_sequences


def execute_with_ddd(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    observable: Optional[Observable] = None,
    *,
    rule: Callable[[int], QPROGRAM],
    rule_args: Dict[str, Any] = {},
    num_trials: int = 1,
    full_output: bool = False,
) -> Union[float, Tuple[float, Dict[str, Any]]]:
    r"""Estimates the error-mitigated expectation value associated to the
    input circuit, via the application of digital dynamical decoupling (DDD).

    Args:
        circuit: The input circuit to execute with DDD.
        executor: A Mitiq executor that executes a circuit and returns the
            unmitigated ``QuantumResult`` (e.g. an expectation value).
        observable: Observable to compute the expectation value of. If None,
            the ``executor`` must return an expectation value. Otherwise,
            the ``QuantumResult`` returned by ``executor`` is used to compute
            the expectation of the observable.
        rule: A function that takes as main argument a slack length (i.e. the
            number of idle moments) of a slack window (i.e. a single-qubit idle
            window in a circuit) and returns the DDD sequence of gates to be
            applied in that window. Mitiq provides standard built-in rules
            that can be directly imported from ``mitiq.ddd.rules``.
        rule_args: An optional dictionary of keyword arguments for ``rule``.
        num_trials: The number of independent experiments to average over.
            A number larger than 1 can be useful to average over multiple
            applications of a rule returning non-deterministic DDD sequences.
        full_output: If ``False`` only the mitigated expectation value is
            returned. If ``True`` a dictionary containing all DDD data is
            returned too.

    Returns:
        The tuple ``(ddd_value, ddd_data)`` where ``ddd_value`` is the
        expectation value estimated with DDD and ``ddd_data`` is a dictionary
        containing all the raw data involved in the DDD process (e.g. the
        circuit filled with DDD sequences). If ``full_output`` is false,
        only ``ddd_value`` is returned.
    """
    # Initialize executor
    if not isinstance(executor, Executor):
        executor = Executor(executor)

    rule_partial: Callable[[int], QPROGRAM]
    rule_partial = partial(rule, **rule_args)

    # Insert DDD sequences in (a copy of) the input circuit
    circuits_with_ddd = [
        insert_ddd_sequences(circuit, rule_partial) for _ in range(num_trials)
    ]
    results = executor.evaluate(
        circuits_with_ddd,
        observable,
        force_run_all=True,
    )

    assert len(results) == num_trials

    ddd_value = np.sum(results) / num_trials
    if not full_output:
        return ddd_value

    ddd_data = {
        "ddd_value": ddd_value,
        "ddd_trials": results,
        "circuits_with_ddd": circuits_with_ddd,
    }
    return ddd_value, ddd_data


def mitigate_executor(
    executor: Callable[[QPROGRAM], QuantumResult],
    observable: Optional[Observable] = None,
    *,
    rule: Callable[[int], QPROGRAM],
    rule_args: Dict[str, Any] = {},
    num_trials: int = 1,
    full_output: bool = False,
) -> Callable[[QPROGRAM], Union[float, Tuple[float, Dict[str, Any]]]]:
    """Returns a modified version of the input 'executor' which is
    error-mitigated with digital dynamical decoupling (DDD).

    Args:
        executor: A function that executes a circuit and returns the
            unmitigated `QuantumResult` (e.g. an expectation value).
        observable: Observable to compute the expectation value of. If None,
            the `executor` must return an expectation value. Otherwise,
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        rule: A function that takes as main argument a slack length (i.e. the
            number of idle moments) of a slack window (i.e. a single-qubit idle
            window in a circuit) and returns the DDD sequence of gates to be
            applied in that window. Mitiq provides standard built-in rules
            that can be directly imported from `mitiq.ddd.rules`.
        rule_args: An optional dictionary of keyword arguments for `rule`.
        num_trials: The number of independent experiments to average over.
            A number larger than 1 can be useful to average over multiple
            applications of a rule returning non-deterministic DDD sequences.
        full_output: If False only the mitigated expectation value is returned.
            If True a dictionary containing all DDD data is returned too.

    Returns:
        The error-mitigated version of the input executor.
    """
    executor_obj = Executor(executor)
    if not executor_obj.can_batch:

        @wraps(executor)
        def new_executor(
            circuit: QPROGRAM,
        ) -> Union[float, Tuple[float, Dict[str, Any]]]:
            return execute_with_ddd(
                circuit,
                executor,
                observable,
                rule=rule,
                rule_args=rule_args,
                num_trials=num_trials,
                full_output=full_output,
            )

    else:

        @wraps(executor)
        def new_executor(
            circuits: List[QPROGRAM],
        ) -> List[Union[float, Tuple[float, Dict[str, Any]]]]:
            return [
                execute_with_ddd(
                    circuit,
                    executor,
                    observable,
                    rule=rule,
                    rule_args=rule_args,
                    num_trials=num_trials,
                    full_output=full_output,
                )
                for circuit in circuits
            ]

    return new_executor


def ddd_decorator(
    observable: Optional[Observable] = None,
    *,
    rule: Callable[[int], QPROGRAM],
    rule_args: Dict[str, Any] = {},
    num_trials: int = 1,
    full_output: bool = False,
) -> Callable[
    [Callable[[QPROGRAM], QuantumResult]],
    Callable[[QPROGRAM], Union[float, Tuple[float, Dict[str, Any]]]],
]:
    """Decorator which adds an error-mitigation layer based on digital
    dynamical decoupling (DDD) to an executor function, i.e., a function which
    executes a quantum circuit with an arbitrary backend and returns a
    ``QuantumResult`` (e.g. an expectation value).

    Args:
        observable: Observable to compute the expectation value of. If None,
            the `executor` must return an expectation value. Otherwise,
            the `QuantumResult` returned by `executor` is used to compute the
            expectation of the observable.
        rule: A function that takes as main argument a slack length (i.e. the
            number of idle moments) of a slack window (i.e. a single-qubit idle
            window in a circuit) and returns the DDD sequence of gates to be
            applied in that window. Mitiq provides standard built-in rules
            that can be directly imported from `mitiq.ddd.rules`.
        rule_args: An optional dictionary of keyword arguments for `rule`.
        num_trials: The number of independent experiments to average over.
            A number larger than 1 can be useful to average over multiple
            applications of a rule returning non-deterministic DDD sequences.
        full_output: If False only the mitigated expectation value is returned.
            If True a dictionary containing all DDD data is returned too.

    Returns:
        The error-mitigating decorator to be applied to an executor function.
    """

    def decorator(
        executor: Callable[[QPROGRAM], QuantumResult],
    ) -> Callable[[QPROGRAM], Union[float, Tuple[float, Dict[str, Any]]]]:
        return mitigate_executor(
            executor,
            observable,
            rule=rule,
            rule_args=rule_args,
            num_trials=num_trials,
            full_output=full_output,
        )

    return decorator



# Copyright (C) Unitary Fund
# Portions of this code have been adapted from PennyLane's tutorial
# on Classical Shadows.
# Original authors: PennyLane developers: Brian Doolittle, Roeland Wiersema
# Tutorial link: https://pennylane.ai/qml/demos/tutorial_classical_shadows
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
"""Classical post-processing process of classical shadows."""

from collections import defaultdict
from functools import reduce
from itertools import compress
from operator import mul
from statistics import median
from typing import Any, Dict, List, Optional, Tuple

import cirq
import numpy as np
import numpy.typing as npt

import mitiq
from mitiq.shadows.shadows_utils import (
    batch_calibration_data,
    create_string,
    valid_bitstrings,
)
from mitiq.utils import matrix_kronecker_product, operator_ptm_vector_rep

# Local unitaries to measure Pauli operators in the Z basis
PAULI_MAP = {
    "X": cirq.unitary(cirq.H),
    "Y": cirq.unitary(cirq.H) @ cirq.unitary(cirq.S).conj(),
    "Z": cirq.unitary(cirq.I),
}

ZERO_STATE = np.array([[1, 0], [0, 0]])
ONE_STATE = np.array([[0, 0], [0, 1]])


def get_single_shot_pauli_fidelity(
    bitstring: str, paulistring: str, locality: Optional[int] = None
) -> Dict[str, float]:
    r"""
    Calculate Pauli fidelity :math:`f_b` for a single shot measurement of the
    calibration circuit for b= bit_string.

    In the notation of arXiv:2011.09636, this function estimates the
    coefficient :math:`f_b`, which characterizes the (noisy) classical
    shadow channel.

    The locality is realized on the assumption that the noisy
    channel :math:`\Lambda` is local
    :math:`\Lambda \equiv \bigotimes_i^n\Lambda_i`.

    Args:
        bit_string: The bitstring corresponding to a computational basis state.
            E.g., '01...0':math:`:=|0\rangle|1\rangle...|0\rangle`.
        pauli_string: The local Pauli measurement performed on each qubit.
            e.g.'XY...Z' means perform local X-basis measurement on the
            1st qubit, local Y-basis measurement the 2ed qubit, local Z-basis
            measurement the last qubit in the circuit.
        locality: The locality of the operator, whose expectation value is
            going to be estimated by the classical shadow. E.g., if the
            operator is the Ising model Hamiltonian with nearest neighbor
            interactions, then locality = 2.

    Returns:
        A dictionary of Pauli fidelity bit_string: :math:`\{{f}_b\}`.
        If the locality is :math:`w < n`, then derive the output's keys from
        the bit_string. Ensure that the number of 1s in the keys is less
        than or equal to w. The corresponding Pauli fidelity is the product of
        local Pauli fidelity where the associated locus in the keys are '1'.
    """
    pauli_fidelity = {"Z0": 1.0, "Z1": -1.0}
    local_fidelities = [
        pauli_fidelity.get(p + b, 0.0) for b, p in zip(bitstring, paulistring)
    ]
    num_qubits = len(bitstring)
    bitstrings = valid_bitstrings(num_qubits, max_hamming_weight=locality)
    fidelities = {}
    for bitstring in bitstrings:
        subset_fidelities = compress(local_fidelities, map(int, bitstring))
        fidelities[bitstring] = reduce(mul, subset_fidelities, 1.0)

    return fidelities


def get_pauli_fidelities(
    calibration_outcomes: Tuple[List[str], List[str]],
    num_batches: int,
    locality: Optional[int] = None,
) -> Dict[str, complex]:
    r"""
    Calculate Pauli fidelities for the calibration circuit. In the notation of
    arXiv:2011.09636, this function estimates the coefficients
    :math:`f_b`, which characterize the (noisy) classical shadow channel
    :math:`\mathcal{M}=\sum_b f_b \Pi_b`.

    Args:
        calibration_measurement_outcomes: The `random_Pauli_measurement`
            outcomes for the state :math:`|0\rangle^{\otimes n}`}` .
        num_batches: The number of batches in the median of means estimator.
        locality: The locality of the operator, whose expectation value is
            going to be estimated by the classical shadow. E.g., if the
            operator is the Ising model Hamiltonian with nearest neighbor
            interactions, then locality = 2.

    Returns:
        A :math:`2^n`-dimensional dictionary of Pauli fidelities
        :math:`f_b` for :math:`b = \{0,1\}^{n}`
    """
    means = defaultdict(list)
    for bitstrings, paulistrings in batch_calibration_data(
        calibration_outcomes, num_batches
    ):
        all_fidelities = defaultdict(list)
        for bitstring, paulistring in zip(bitstrings, paulistrings):
            fidelities = get_single_shot_pauli_fidelity(
                bitstring, paulistring, locality=locality
            )
            for b, f in fidelities.items():
                all_fidelities[b].append(f)

        for bitstring, fids in all_fidelities.items():
            means[bitstring].append(sum(fids) / num_batches)

    return {
        bitstring: median(averages) for bitstring, averages in means.items()
    }


def classical_snapshot(
    bitstring: str,
    paulistring: str,
    fidelities: Optional[Dict[str, float]] = None,
) -> npt.NDArray[Any]:
    r"""
    Implement a single snapshot state reconstruction
    with calibration of the noisy quantum channel.

    Args:
        bitstring: A bitstring corresponding to the outcome ... TODO
        paulistring: String of the applied Pauli measurement on each qubit.
        f_est: The estimated Pauli fidelities to use for calibration if
            available.

    Returns:
        Reconstructed classical snapshot in terms of nparray.
    """
    # calibrate the noisy quantum channel, output in PTM rep.
    # ptm rep of identity
    I_ptm = operator_ptm_vector_rep(np.eye(2) / np.sqrt(2))
    pi_zero = np.outer(I_ptm, I_ptm)
    pi_one = np.eye(4) - pi_zero
    pi_zero = np.diag(pi_zero)
    pi_one = np.diag(pi_one)

    if fidelities:
        elements = []
        for bits, fidelity in fidelities.items():
            pi_snapshot_vector = []
            for b1, b2, pauli in zip(bits, bitstring, paulistring):
                # get pi for each qubit based on calibration measurement
                pi = pi_zero if b1 == "0" else pi_one
                # get state for each qubit based on shadow measurement
                state = ZERO_STATE if b2 == "0" else ONE_STATE
                # get U for each qubit based on shadow measurement
                U = PAULI_MAP[pauli]
                pi_snapshot_vector.append(
                    pi * operator_ptm_vector_rep(U.conj().T @ state @ U)
                )
            elements.append(
                1 / fidelity * matrix_kronecker_product(pi_snapshot_vector)
            )
        rho_snapshot = np.sum(elements, axis=0)
    else:
        local_rhos = []
        for bit, pauli in zip(bitstring, paulistring):
            state = ZERO_STATE if bit == "0" else ONE_STATE
            U = PAULI_MAP[pauli]
            # apply inverse of the quantum channel,get PTM vector rep
            local_rho = 3.0 * (U.conj().T @ state @ U) - cirq.unitary(cirq.I)
            local_rhos.append(local_rho)

        rho_snapshot = matrix_kronecker_product(local_rhos)
    return rho_snapshot


def shadow_state_reconstruction(
    shadow_measurement_outcomes: Tuple[List[str], List[str]],
    fidelities: Optional[Dict[str, float]] = None,
) -> npt.NDArray[Any]:
    """Reconstruct a state approximation as an average over all snapshots.

    Args:
        shadow_measurement_outcomes: Measurement result and the basis
            performing the measurement obtained from `random_pauli_measurement`
            for classical shadow protocol.
        f_est: The estimated Pauli fidelities to use for calibration if
            available.
    Returns:
        The state reconstructed from classical shadow protocol
    """
    bitstrings, paulistrings = shadow_measurement_outcomes

    return np.mean(
        [
            classical_snapshot(bitstring, paulistring, fidelities)
            for bitstring, paulistring in zip(bitstrings, paulistrings)
        ],
        axis=0,
    )


def expectation_estimation_shadow(
    measurement_outcomes: Tuple[List[str], List[str]],
    pauli: mitiq.PauliString,
    num_batches: int,
    fidelities: Optional[Dict[str, float]] = None,
) -> float:
    """Calculate the expectation value of an observable from classical shadows.
    Use median of means to ameliorate the effects of outliers.

    Args:
        measurement_outcomes: A shadow tuple obtained from
            `z_basis_measurement`.
        pauli_str: Single mitiq observable consisting of
            Pauli operators.
        num_batches: Number of batches to process measurement outcomes in.
        f_est: The estimated Pauli fidelities to use for calibration if
            available.

    Returns:
        Float corresponding to the estimate of the observable expectation
        value.
    """
    bitstrings, paulistrings = measurement_outcomes
    num_qubits = len(bitstrings[0])

    qubits = sorted(pauli.support())
    filtered_bitstrings = [
        "".join([bitstring[q] for q in qubits]) for bitstring in bitstrings
    ]
    filtered_paulis = [
        "".join([pauli[q] for q in qubits]) for pauli in paulistrings
    ]
    filtered_data = (filtered_bitstrings, filtered_paulis)

    means = []
    for bits, paulis in batch_calibration_data(filtered_data, num_batches):
        matching_indices = [i for i, p in enumerate(paulis) if p == pauli.spec]
        if matching_indices:
            matching_bits = (bits[i] for i in matching_indices)
            product = sum((-1) ** bit.count("1") for bit in matching_bits)

            if fidelities:
                b = create_string(num_qubits, qubits)
                product /= fidelities.get(b, np.inf)
            else:
                product *= 3 ** len(qubits)

        else:
            product = 0.0

        means.append(product / len(bits))

    return np.real(np.median(means) * pauli.coeff)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.interface.mitiq_qibo.conversions import (
    from_qibo,
    to_qibo,
    UnsupportedQiboCircuitError,
)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions to create a QPE circuit."""

from typing import Optional

import cirq

from mitiq import QPROGRAM
from mitiq.interface import convert_from_mitiq


def generate_qpe_circuit(
    evalue_reg: int,
    input_gate: cirq.Gate = cirq.T,
    return_type: Optional[str] = None,
) -> QPROGRAM:
    """Returns a circuit to create a quantum phase estimation (QPE) circuit as
    defined in https://en.wikipedia.org/wiki/Quantum_phase_estimation_algorithm

    The unitary to estimate the phase of corresponds to a
    single-qubit gate (``input_gate``).

    The IQFT circuit defined in this method is taken from taken from Sec 7.7.4
    of :cite:`Wong_2022`. The notation for eigenvalue register and eigenstate
    register used to define this function also follows from :cite:`Wong_2022`.

    Args:
        evalue_reg : Number of qubits in the eigenvalue register. The qubits
            in this variable are used to estimate the phase.
        input_gate : The unitary to estimate the phase of as a single-qubit
            Cirq gate. Default gate used here is `cirq.T`.
        return_type: Return type of the output circuit.
    Returns:
        A Quantum Phase Estimation circuit.
    """
    if evalue_reg <= 0:
        raise ValueError(
            "{} is invalid for the number of eigenvalue reg qubits. ",
            evalue_reg,
        )
    num_qubits_for_gate = input_gate.num_qubits()
    if num_qubits_for_gate > 1:
        raise ValueError("This QPE method only works for 1-qubit gates.")

    if evalue_reg == num_qubits_for_gate:
        raise ValueError(
            "The eigenvalue reg must be larger than the eigenstate reg."
        )

    total_num_qubits = evalue_reg + num_qubits_for_gate
    qreg = cirq.LineQubit.range(total_num_qubits)
    circuit = cirq.Circuit()

    # QFT circuit
    # apply hadamard and controlled unitary to the qubits in the eigenvalue reg
    hadamard_circuit = cirq.Circuit()
    for i in range(evalue_reg):
        hadamard_circuit.append(cirq.H(qreg[i]))
    circuit = circuit + hadamard_circuit

    for i in range(total_num_qubits - 1)[::-1]:
        circuit.append(
            [input_gate(qreg[-1]).controlled_by(qreg[i])]
            * (2 ** (evalue_reg - 1 - i))
        )

    # IQFT of the eigenvalue register
    # swap the qubits in the eigenvalue register
    for i in range(int(evalue_reg / 2)):
        circuit.append(
            cirq.SWAP(qreg[i], qreg[evalue_reg - 1 - i]),
            strategy=cirq.InsertStrategy.NEW,
        )
    # apply inverse of hadamard followed by controlled unitary
    circuit.append(cirq.H(qreg[0]), strategy=cirq.InsertStrategy.NEW)
    for i in range(1, evalue_reg):
        for j in range(evalue_reg):
            if j < i:
                circuit.append(
                    cirq.inverse(input_gate(qreg[i]).controlled_by(qreg[j])),
                    strategy=cirq.InsertStrategy.NEW,
                )
        circuit.append(
            cirq.H(qreg[i]),
            strategy=cirq.InsertStrategy.NEW,
        )
    return_type = "cirq" if not return_type else return_type

    return convert_from_mitiq(circuit, return_type)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

import copy
from collections import defaultdict
from numbers import Number
from typing import Any, Callable, Iterable, List, Optional, Set, Union, cast

import cirq
import numpy as np
import numpy.typing as npt

from mitiq import QPROGRAM, MeasurementResult, QuantumResult
from mitiq.observable.pauli import PauliString, PauliStringCollection


class Observable:
    """A quantum observable typically used to compute its mitigated expectation
    value.

    """

    def __init__(self, *paulis: PauliString) -> None:
        """Initializes an `Observable` with :class:`.PauliString` objects.

        Args:
            paulis: PauliStrings used to define the observable.

        """
        self._paulis = _combine_duplicate_pauli_strings(paulis)
        self._groups: List[PauliStringCollection]
        self._ngroups: int
        self.partition()

    @staticmethod
    def from_pauli_string_collections(
        *pauli_string_collections: PauliStringCollection,
    ) -> "Observable":
        obs = Observable()
        obs._groups = list(pauli_string_collections)
        obs._ngroups = len(pauli_string_collections)
        obs._paulis = [
            pauli
            for pauli_string_collection in pauli_string_collections
            for pauli in pauli_string_collection.elements
        ]
        return obs

    @property
    def nterms(self) -> int:
        return len(self._paulis)

    def _qubits(self) -> Set[cirq.Qid]:
        """Returns all qubits acted on by the Observable."""
        return {q for pauli in self._paulis for q in pauli._pauli.qubits}

    @property
    def paulis(self) -> List[PauliString]:
        return self._paulis

    @property
    def qubit_indices(self) -> List[int]:
        return [cast(cirq.LineQubit, q).x for q in sorted(self._qubits())]

    @property
    def nqubits(self) -> int:
        return len(self.qubit_indices)

    def __mul__(
        self, other: Union["Observable", "PauliString", Number]
    ) -> "Observable":
        if isinstance(other, (PauliString, Number)):
            return Observable(*[pauli * other for pauli in self._paulis])
        elif isinstance(other, Observable):
            return Observable(
                *[
                    pauli * other_pauli
                    for pauli in self._paulis
                    for other_pauli in other._paulis
                ]
            )
        return NotImplemented

    def __rmul__(self, other: Union["PauliString", Number]) -> "Observable":
        if isinstance(other, (PauliString, Number)):
            return Observable(*[other * pauli for pauli in self._paulis])
        return NotImplemented

    @property
    def groups(self) -> List[PauliStringCollection]:
        return self._groups

    @property
    def ngroups(self) -> int:
        return self._ngroups

    def partition(self, seed: Optional[int] = None) -> None:
        rng = np.random.RandomState(seed)

        psets: List[PauliStringCollection] = []
        paulis = copy.deepcopy(self._paulis)
        rng.shuffle(paulis)  # type: ignore

        while paulis:
            pauli = paulis.pop()
            added = False
            for i, pset in enumerate(psets):
                if pset.can_add(pauli):
                    pset.add(pauli)
                    added = True
                    break

            if not added:
                psets.append(PauliStringCollection(pauli))

        self._groups = psets
        self._ngroups = len(self._groups)

    def measure_in(self, circuit: QPROGRAM) -> List[QPROGRAM]:
        return [pset.measure_in(circuit) for pset in self._groups]

    def matrix(
        self,
        qubit_indices: Optional[List[int]] = None,
    ) -> npt.NDArray[np.complex64]:
        """Returns the (potentially very large) matrix of the Observable."""
        if qubit_indices is None:
            qubit_indices = self.qubit_indices
        n = len(qubit_indices)

        obs_matrix = np.zeros(shape=(2**n, 2**n), dtype=np.complex64)
        for pauli in self._paulis:
            obs_matrix += pauli.matrix(qubit_indices_to_include=qubit_indices)

        return obs_matrix

    def expectation(
        self, circuit: QPROGRAM, execute: Callable[[QPROGRAM], QuantumResult]
    ) -> complex:
        from mitiq.executor import Executor  # Avoid circular import.

        return Executor(execute).evaluate(circuit, observable=self)[0]

    def _expectation_from_measurements(
        self, measurements: List[MeasurementResult]
    ) -> float:
        return sum(
            pset._expectation_from_measurements(bitstrings)
            for (pset, bitstrings) in zip(self._groups, measurements)
        )

    def _expectation_from_density_matrix(
        self, density_matrix: npt.NDArray[np.complex64]
    ) -> float:
        observable_matrix = self.matrix()

        if density_matrix.shape != observable_matrix.shape:
            nqubits = int(np.log2(density_matrix.shape[0]))
            density_matrix = cirq.partial_trace(
                np.reshape(density_matrix, newshape=[2, 2] * nqubits),
                keep_indices=self.qubit_indices,
            ).reshape(observable_matrix.shape)

        return np.real_if_close(
            np.trace(density_matrix @ observable_matrix)
        ).item()

    def __str__(self) -> str:
        return " + ".join(map(str, self._paulis))

    def __eq__(self, other: Any) -> bool:
        return np.allclose(self.matrix(), other.matrix())


def _combine_duplicate_pauli_strings(
    paulis: Iterable[PauliString],
) -> List[PauliString]:
    """Combines duplicate PauliStrings by adding their coefficients.
    Discards paulis with zero coefficients.

    Returns: deduped list of PauliStrings.
    """
    pauli_string_coefficients: defaultdict[PauliString, complex] = defaultdict(
        complex
    )
    for pauli_string in paulis:
        cache_key = pauli_string.with_coeff(1)
        pauli_string_coefficients[cache_key] += pauli_string.coeff
    return [
        pauli_string.with_coeff(coeff)
        for (pauli_string, coeff) in pauli_string_coefficients.items()
        if not np.isclose(coeff, 0.0)
    ]



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Tools to determine slack windows in circuits and to insert DDD sequences."""

from typing import Callable

import numpy as np
import numpy.typing as npt
from cirq import Circuit, I, LineQubit, synchronize_terminal_measurements

from mitiq import QPROGRAM
from mitiq.interface import accept_qprogram_and_validate


def _get_circuit_mask(circuit: Circuit) -> npt.NDArray[np.int64]:
    """Given a circuit with n qubits and d moments returns a matrix
    :math:`A` with n rows and d columns. The matrix elements are
    :math:`A_{i,j} = 1` if there is a non-identity gate acting on qubit
    :math:`i` at moment :math:`j`, while :math:`A_{i,j} = 0` otherwise.

    Args:
        circuit: Input circuit to mask with n qubits and d moments

    Returns:
        A mask matrix with n rows and d columns
    """
    qubits = sorted(circuit.all_qubits())
    indexed_qubits = [(i, n) for (i, n) in enumerate(qubits)]
    mask_matrix = np.zeros((len(qubits), len(circuit)), dtype=int)
    for moment_index, moment in enumerate(circuit):
        for op in moment:
            qubit_indices = [
                qubit[0]
                for qubit in indexed_qubits
                if qubit[1] in op.qubits and op.gate != I
            ]
            for qubit_index in qubit_indices:
                mask_matrix[qubit_index, moment_index] = 1
    return mask_matrix


def _validate_integer_matrix(mask: npt.NDArray[np.int64]) -> None:
    """Ensures the input is a NumPy 2d array with integer elements."""
    if not isinstance(mask, np.ndarray):
        raise TypeError("The input matrix must be a numpy.ndarray object.")
    if not np.issubdtype(mask.dtype.type, int):
        raise TypeError("The input matrix must have integer elements.")
    if len(mask.shape) != 2:
        raise ValueError("The input must be a 2-dimensional array.")


def get_slack_matrix_from_circuit_mask(
    mask: npt.NDArray[np.int64],
) -> npt.NDArray[np.int64]:
    """Given a circuit mask matrix :math:`A`, e.g., the output of
    ``_get_circuit_mask()``, returns a slack matrix :math:`B`,
    where :math:`B_{i,j} = t` if the position :math:`A_{i,j}` is the
    initial element of a sequence of :math:`t` zeros (from left to right).

    Args:
        mask: The mask matrix of a quantum circuit.

    Returns:
        The matrix of slack lengths.
    """
    _validate_integer_matrix(mask)
    if not (mask**2 == mask).all():
        raise ValueError("The input matrix elements must be 0 or 1.")

    num_rows, num_cols = mask.shape
    slack_matrix = np.zeros((num_rows, num_cols), dtype=int)
    for r in range(num_rows):
        for c in range(num_cols):
            previous_elem = mask[r, c - 1] if c != 0 else 1
            if previous_elem == 1:
                # Compute slack length
                for elem in mask[r, c::]:
                    if elem == 0:
                        slack_matrix[r, c] += 1
                    else:
                        break

    return slack_matrix


def insert_ddd_sequences(
    circuit: QPROGRAM,
    rule: Callable[[int], Circuit],
) -> QPROGRAM:
    """Returns the circuit with DDD sequences applied according to the input
    rule.

    Args:
        circuit: The QPROGRAM circuit to be modified with DDD sequences.
        rule: The rule determining what DDD sequences should be applied.
            A set of built-in DDD rules can be imported from
            ``mitiq.ddd.rules``.

    Returns:
        The circuit with DDD sequences added.
    """

    return _insert_ddd_sequences(circuit, rule)


@accept_qprogram_and_validate
def _insert_ddd_sequences(
    circuit: Circuit,
    rule: Callable[[int], Circuit],
) -> Circuit:
    """Returns the circuit with DDD sequences applied according to the input
    rule.

    Args:
        circuit: The Cirq circuit to be modified with DDD sequences.
        rule: The rule determining what DDD sequences should be applied.
            A set of built-in DDD rules can be imported from
            ``mitiq.ddd.rules``.

    Returns:
        The circuit with DDD sequences added.
    """
    circuit = synchronize_terminal_measurements(circuit)
    if not circuit.are_all_measurements_terminal():
        raise ValueError(
            "This circuit contains midcircuit measurements which "
            "are not currently supported by DDD."
        )

    slack_matrix = get_slack_matrix_from_circuit_mask(
        _get_circuit_mask(circuit)
    )
    # Copy to avoid mutating the input circuit
    circuit_with_ddd = circuit.copy()
    qubits = sorted(circuit.all_qubits())
    for moment_idx in range(len(circuit)):
        slack_column = slack_matrix[:, moment_idx]
        for row_index, slack_length in enumerate(slack_column):
            if slack_length > 1:
                ddd_sequence = rule(slack_length).transform_qubits(
                    {LineQubit(0): qubits[row_index]}
                )
                for idx, op in enumerate(ddd_sequence.all_operations()):
                    moment = circuit_with_ddd[moment_idx + idx]
                    op_to_replace = moment.operation_at(*op.qubits)

                    if op_to_replace and op_to_replace.gate == I:
                        moment = moment.without_operations_touching(op.qubits)

                    circuit_with_ddd[moment_idx + idx] = moment.with_operation(
                        op
                    )
    return circuit_with_ddd



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Built-in rules determining what DDD sequence should be applied in a given
slack window.
"""

from itertools import cycle
from typing import List

import numpy as np
from cirq import (
    Circuit,
    Gate,
    I,
    LineQubit,
    X,
    Y,
    allclose_up_to_global_phase,
    unitary,
)


def general_rule(
    slack_length: int, gates: List[Gate], spacing: int = -1
) -> Circuit:
    """Returns a digital dynamical decoupling sequence, based on inputs.

    Args:
        slack_length: Length of idle window to fill.
        spacing: How many identity spacing gates to apply between dynamical
            decoupling gates, as a non-negative int. Negative int corresponds
            to default. Defaults to maximal spacing that fits a single sequence
            in the given slack window.
        gates: A list of single qubit Cirq gates to build the rule. E.g. [X, X]
            is the xx sequence, [X, Y, X, Y] is the xyxy sequence.
            - Note: To repeat the sequence, specify a repeated gateset.
    Returns:
        A digital dynamical decoupling sequence, as a Cirq circuit.

    Example:
        When ``slack_length = 8`` and ``gates = [X, X]`` the spacing defaults
        to 2 and the rule returns the sequence::

            ──I──I──X──I──I──X──I──I──

        When ``slack_length = 9`` and ``gates [X, Y, X, Y]`` the spacing
        defaults to 1 and the rule returns the sequence::

            ──I──X──I──Y──I──X──I──Y──I──
    """
    if len(gates) < 2:
        raise ValueError("Gateset too short to make a ddd sequence.")
    if slack_length < 2 or slack_length < len(gates):
        return Circuit()
    num_decoupling_gates = len(gates)
    slack_difference = slack_length - num_decoupling_gates
    if spacing < 0:
        spacing = slack_difference // (num_decoupling_gates + 1)
    slack_remainder = slack_length - (
        spacing * (num_decoupling_gates + 1) + num_decoupling_gates
    )
    if slack_remainder < 0:
        return Circuit()
    q = LineQubit(0)
    slack_gates = [I(q) for _ in range(spacing)]
    sequence = Circuit(slack_gates)
    for gate in gates:
        sequence.append(gate.on(q))
        sequence.append(slack_gates)

    if not allclose_up_to_global_phase(np.eye(2), unitary(sequence)):
        raise ValueError("Sequence is not equivalent to the identity!")
    for i in range(slack_remainder):
        sequence.append(I(q)) if i % 2 else sequence.insert(0, I(q))
    return sequence


def xx(slack_length: int, spacing: int = -1) -> Circuit:
    """Returns an XX digital dynamical decoupling sequence, based on inputs.

    Args:
        slack_length: Length of idle window to fill.
        spacing: How many identity spacing gates to apply between dynamical
            decoupling gates, as a non-negative int. Negative int corresponds
            to default. Defaults to maximal spacing that fits a single sequence
            in the given slack window.
            E.g. given slack_length = 8 the spacing defaults to 2 and this
            rule returns the sequence:
            ──I──I──X──I──I──X──I──I──.
    Returns:
        An XX digital dynamical decoupling sequence, as a Cirq circuit.
    """
    xx_sequence = general_rule(
        slack_length=slack_length,
        spacing=spacing,
        gates=[X, X],
    )
    return xx_sequence


def xyxy(slack_length: int, spacing: int = -1) -> Circuit:
    """Returns an XYXY digital dynamical decoupling sequence, based on inputs.

    Args:
        slack_length: Length of idle window to fill.
        spacing: How many identity spacing gates to apply between dynamical
            decoupling gates, as a non-negative int. Negative int corresponds
            to default. Defaults to maximal spacing that fits a single sequence
            in the given slack window.
            E.g. given slack_length = 9 the spacing defaults to 1 and this
            rule returns the sequence:
            ──I──X──I──Y──I──X──I──Y──I──.
    Returns:
        An XYXY digital dynamical decoupling sequence, as a Cirq circuit.
    """
    xyxy_sequence = general_rule(
        slack_length=slack_length,
        spacing=spacing,
        gates=[X, Y, X, Y],
    )
    return xyxy_sequence


def yy(slack_length: int, spacing: int = -1) -> Circuit:
    """Returns a YY digital dynamical decoupling sequence, based on inputs.

    Args:
        slack_length: Length of idle window to fill.
        spacing: How many identity spacing gates to apply between dynamical
            decoupling gates, as a non-negative int. Negative int corresponds
            to default. Defaults to maximal spacing that fits a single sequence
            in the given slack window.
            E.g. given slack_length = 8 the spacing defaults to 2 and
            this rule returns the sequence:
            ──I──I──Y──I──I──Y──I──I──.
    Returns:
        An YY digital dynamical decoupling sequence, as a Cirq circuit.
    """
    yy_sequence = general_rule(
        slack_length=slack_length,
        spacing=spacing,
        gates=[Y, Y],
    )
    return yy_sequence


def repeated_rule(slack_length: int, gates: List[Gate]) -> Circuit:
    """Returns a general digital dynamical decoupling sequence that repeats
    until the slack is filled without spacing, up to a complete repetition.

    Args:
        slack_length: Length of idle window to fill.
        gates: A list of single qubit Cirq gates to build the rule. E.g. [X, X]
            is the xx sequence, [X, Y, X, Y] is the xyxy sequence.
    Returns:
        A repeated digital dynamical decoupling sequence, as a Cirq circuit.

    Warning:
        Where :func:`.general_rule()` fills a slack window with a single
        sequence, this rule attempts to fill every moment with sequence
        repetitions (up to a complete repetition of the gate set).
        E.g. given ``slack_length = 8`` and ``gates = [X, Y, X, Y]``, this
        rule returns the sequence::

            ──X──Y──X──Y──X──Y──X──Y──
    """
    num_decoupling_gates = len(gates)
    if num_decoupling_gates > slack_length:
        return Circuit()
    sequence = general_rule(
        slack_length=slack_length,
        spacing=0,
        gates=[
            gate
            for (_, gate) in zip(
                range(slack_length - (slack_length % num_decoupling_gates)),
                cycle(gates),
            )
        ],
    )
    return sequence



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Defines utilities for efficiently running collections of circuits generated
by error mitigation techniques to compute expectation values."""

import inspect
import warnings
from collections import Counter
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

import numpy as np
import numpy.typing as npt

from mitiq import QPROGRAM, MeasurementResult, QuantumResult
from mitiq.interface import convert_from_mitiq, convert_to_mitiq
from mitiq.observable.observable import Observable
from mitiq.observable.pauli import PauliString

DensityMatrixLike = [
    np.ndarray,
    Iterable[np.ndarray],  # type: ignore
    List[np.ndarray],  # type: ignore
    Sequence[np.ndarray],  # type: ignore
    Tuple[np.ndarray],
    npt.NDArray[np.complex64],
]
FloatLike = [
    None,  # Untyped executors are assumed to return floats.
    float,
    Iterable[float],
    List[float],
    Sequence[float],
    Tuple[float],
]
MeasurementResultLike = [
    MeasurementResult,
    Iterable[MeasurementResult],
    List[MeasurementResult],
    Sequence[MeasurementResult],
    Tuple[MeasurementResult],
]


class Executor:
    """Tool for efficiently scheduling/executing quantum programs and storing
    the results.
    """

    def __init__(
        self,
        executor: Callable[[Union[QPROGRAM, Sequence[QPROGRAM]]], Any],
        max_batch_size: int = 75,
    ) -> None:
        """Initializes an Executor.

        Args:
            executor: A function which inputs a program and outputs a
                ``mitiq.QuantumResult``, or inputs a sequence of programs and
                outputs a sequence of ``mitiq.QuantumResult`` s.
            max_batch_size: Maximum number of programs that can be sent in a
                single batch (if the executor is batched).
        """
        self._executor = executor

        executor_annotation = inspect.getfullargspec(executor).annotations
        self._executor_return_type = executor_annotation.get("return")
        self._max_batch_size = max_batch_size

        self._executed_circuits: List[QPROGRAM] = []
        self._quantum_results: List[QuantumResult] = []

        self._calls_to_executor: int = 0

    @property
    def can_batch(self) -> bool:
        return self._executor_return_type in (
            BatchedType[T]  # type: ignore[index]
            for BatchedType in [Iterable, List, Sequence, Tuple]
            for T in QuantumResult.__args__  # type: ignore[attr-defined]
        )

    @property
    def executed_circuits(self) -> List[QPROGRAM]:
        return self._executed_circuits

    @property
    def quantum_results(self) -> List[QuantumResult]:
        return self._quantum_results

    @property
    def calls_to_executor(self) -> int:
        return self._calls_to_executor

    def evaluate(
        self,
        circuits: Union[QPROGRAM, List[QPROGRAM]],
        observable: Optional[Observable] = None,
        force_run_all: bool = True,
        **kwargs: Any,
    ) -> List[float]:
        """Returns the expectation value Tr[ρ O] for each circuit in
        ``circuits`` where O is the observable provided or implicitly defined
        by the ``executor``. (The observable is implicitly defined when the
        ``executor`` returns float(s).)

        All executed circuits are stored in ``self.executed_circuits``, and all
        quantum results are stored in ``self.quantum_results``.

        Args:
            circuits: A single circuit of list of circuits.
            observable: Observable O in the expression Tr[ρ O]. If None,
                the ``executor`` must return a float (which corresponds to
                Tr[ρ O] for a specific, fixed observable O).
            force_run_all: If True, force every circuit in the input sequence
                to be executed (if some are identical). Else, detects identical
                circuits and runs a minimal set.

        Returns:
            List of real valued expectation values.
        """
        if not isinstance(circuits, List):
            circuits = [circuits]

        warn_non_hermitian = False
        if observable:
            if isinstance(observable, PauliString):
                if observable.coeff.imag > 0.0001:
                    warn_non_hermitian = True
            elif isinstance(observable, Observable):
                if any(
                    pauli.coeff.imag > 0.0001 for pauli in observable._paulis
                ):
                    warn_non_hermitian = True
        if warn_non_hermitian:
            warnings.warn(
                "Expected observable to be hermitian. Continue with caution."
            )

        # Get all required circuits to run.
        if (
            observable is not None
            and self._executor_return_type in MeasurementResultLike
        ):
            all_circuits = [
                circuit_with_measurements
                for circuit in circuits
                for circuit_with_measurements in observable.measure_in(circuit)
            ]
            result_step = observable.ngroups
        elif (
            observable is not None
            and self._executor_return_type not in MeasurementResultLike
            and self._executor_return_type not in DensityMatrixLike
        ):
            raise ValueError(
                """Executor and observable are not compatible. Executors
                returning expectation values as float must be used with
                observable=None"""
            )
        else:
            all_circuits = circuits
            result_step = 1

        # Run all required circuits.
        all_results = self.run(all_circuits, force_run_all, **kwargs)

        # Parse the results.
        if self._executor_return_type in FloatLike:
            results = np.real_if_close(
                cast(Sequence[float], all_results)
            ).tolist()

        elif self._executor_return_type in DensityMatrixLike:
            observable = cast(Observable, observable)
            all_results = cast(List[npt.NDArray[np.complex64]], all_results)
            results = [
                observable._expectation_from_density_matrix(density_matrix)
                for density_matrix in all_results
            ]

        elif self._executor_return_type in MeasurementResultLike:
            observable = cast(Observable, observable)
            all_results = cast(List[MeasurementResult], all_results)
            results = [
                observable._expectation_from_measurements(
                    all_results[i : i + result_step]
                )
                for i in range(len(all_results) // result_step)
            ]

        else:
            raise ValueError(
                f"Could not parse executed results from executor with type"
                f" {self._executor_return_type}."
            )

        return results

    def run(
        self,
        circuits: Union[QPROGRAM, Sequence[QPROGRAM]],
        force_run_all: bool = True,
        **kwargs: Any,
    ) -> Sequence[QuantumResult]:
        """Runs all input circuits using the least number of possible calls to
        the executor.

        Args:
            circuits: Circuit or sequence thereof to execute with the executor.
            force_run_all: If True, force every circuit in the input sequence
                to be executed (if some are identical). Else, detects identical
                circuits and runs a minimal set.
        """
        if not isinstance(circuits, Sequence):
            circuits = [circuits]

        start_result_index = len(self._quantum_results)

        if force_run_all:
            to_run = circuits
        else:
            # Make circuits hashable.
            # Note: Assumes all circuits are the same type.
            # TODO: Bug! These conversions to/from Mitiq are not safe in that,
            #  e.g., they do not preserve classical register structure in
            #  Qiskit circuits, potentially causing executed results to be
            #  incorrect. Safe conversions should follow the logic in
            #  mitiq.interface.noise_scaling_converter.
            _, conversion_type = convert_to_mitiq(circuits[0])
            hashable_circuits = [
                convert_to_mitiq(circ)[0].freeze() for circ in circuits
            ]

            # Get the unique circuits and counts
            collection = Counter(hashable_circuits)
            to_run = [
                convert_from_mitiq(circ.unfreeze(), conversion_type)
                for circ in collection.keys()
            ]

        if not self.can_batch:
            for circuit in to_run:
                self._call_executor(circuit, **kwargs)

        else:
            stop = len(to_run)
            step = self._max_batch_size
            for i in range(int(np.ceil(stop / step))):
                batch = to_run[i * step : (i + 1) * step]
                self._call_executor(batch, **kwargs)

        results = self._quantum_results[start_result_index:]

        if not force_run_all:
            # Expand computed results to all results using counts.
            results_dict = dict(zip(collection.keys(), results))
            results = [results_dict[key] for key in hashable_circuits]

        return self._post_run(results)

    def _post_run(
        self, results: Sequence[QuantumResult]
    ) -> Sequence[QuantumResult]:
        """Post-processes the measurement results.
        For example, this method can be overridden by a
        readout error mitigation function.
        """
        return results

    def _call_executor(
        self, to_run: Union[QPROGRAM, Sequence[QPROGRAM]], **kwargs: Any
    ) -> None:
        """Calls the executor on the input circuit(s) to run. Stores the
        executed circuits in ``self._executed_circuits`` and the quantum
        results in ``self._quantum_results``.

        Args:
            to_run: Circuit(s) to run.
        """
        result = self._executor(to_run, **kwargs)
        self._calls_to_executor += 1

        if self.can_batch:
            self._quantum_results.extend(result)
            self._executed_circuits.extend(to_run)
        else:
            self._quantum_results.append(result)
            self._executed_circuits.append(to_run)

    @staticmethod
    def is_batched_executor(
        executor: Callable[[Union[QPROGRAM, Sequence[QPROGRAM]]], Any],
    ) -> bool:
        """Returns True if the input function is recognized as a "batched
        executor", else False.

        The executor is detected as "batched" if and only if it is annotated
        with a return type that is one of the following:

        * ``Iterable[QuantumResult]``
        * ``List[QuantumResult]``
        * ``Sequence[QuantumResult]``
        * ``Tuple[QuantumResult]``

        Otherwise, it is considered "serial".

        Batched executors can _run several quantum programs in a single call.
        See below.

        Args:
            executor: A "serial executor" (1) or a "batched executor" (2).

                #. A function which inputs a single ``QPROGRAM`` and outputs a
                   single ``QuantumResult``.
                #. A function which inputs a list of ``QPROGRAM`` objects and
                   returns a list of ``QuantumResult`` instances (one for each
                   ``QPROGRAM``).

        Returns:
            True if the executor is detected as batched, else False.
        """
        executor_annotation = inspect.getfullargspec(executor).annotations

        return executor_annotation.get("return") in (
            BatchedType[T]  # type: ignore[index]
            for BatchedType in [Iterable, List, Sequence, Tuple]
            for T in QuantumResult.__args__  # type: ignore[attr-defined]
        )



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions for computing the projector for subspace expansion."""

from itertools import product
from typing import Callable, Dict, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
from numpy.linalg import pinv
from scipy.linalg import eigh

from mitiq import QPROGRAM, Executor, Observable, PauliString, QuantumResult


def get_projector(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    check_operators: Sequence[PauliString],
    code_hamiltonian: Observable,
    pauli_string_to_expectation_cache: Dict[PauliString, complex] = {},
) -> Observable:
    """Computes the projector onto the code space defined by the
    check_operators provided that minimizes the code_hamiltonian.

    Returns: Projector as an Observable.
    """
    S = _compute_overlap_matrix(
        circuit, executor, check_operators, pauli_string_to_expectation_cache
    )
    H = _compute_overlap_matrix(
        circuit,
        executor,
        check_operators,
        pauli_string_to_expectation_cache,
        code_hamiltonian,
    )
    # We only want the smallest eigenvalue and corresponding eigenvector
    _, C = eigh(pinv(S) @ H, subset_by_index=[0, 0])
    # np float type: np.float64

    Cs = C[:, 0]
    projector = Observable(
        *[check_operators[i] * Cs[i] for i in range(len(check_operators))]
    )

    return projector


def get_expectation_value_for_observable(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    observable: Union[PauliString, Observable],
    pauli_expectation_cache: Dict[PauliString, complex] = {},
) -> float:
    """Provide pauli_string_to_expectation_cache if you want to take advantage
    of caching.

    This function modifies pauli_string_to_expectation_cache in place.
    """

    final_executor = (
        executor if isinstance(executor, Executor) else Executor(executor)
    )

    def get_expectation_value_for_one_pauli(
        pauli_string: PauliString,
    ) -> float:
        cache_key = pauli_string.with_coeff(1)
        pauli_expectation_cache[cache_key] = final_executor.evaluate(
            circuit, Observable(cache_key)
        )[0]
        return (pauli_expectation_cache[cache_key] * pauli_string.coeff).real

    paulis = (
        [observable]
        if isinstance(observable, PauliString)
        else observable.paulis
    )
    expectation_value = sum(
        get_expectation_value_for_one_pauli(pauli) for pauli in paulis
    )
    return expectation_value


def _compute_overlap_matrix(
    circuit: QPROGRAM,
    executor: Union[Executor, Callable[[QPROGRAM], QuantumResult]],
    check_operators: Sequence[PauliString],
    pauli_expectation_cache: Dict[PauliString, complex] = {},
    code_hamiltonian: Optional[Observable] = None,
) -> npt.NDArray[np.float64]:
    num_ops = len(check_operators)

    H = np.zeros((num_ops, num_ops))
    # Hij = ⟨Ψ|Mi† H Mj|Ψ⟩
    for i, j in product(range(num_ops), repeat=2):
        observable: Union[PauliString, Observable]
        if code_hamiltonian:
            observable = (
                check_operators[i] * code_hamiltonian * check_operators[j]
            )
        else:
            observable = check_operators[i] * check_operators[j]
        H[i, j] = get_expectation_value_for_observable(
            circuit, executor, observable, pauli_expectation_cache
        )
    return H



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
#
# This module includes a modified version of the ApplyLayout class originally
# licensed under the Apache License, Version 2.0.
#
# Original Author: IBM
# Modification made by: André Alves on 07 Apr 2024
#
# For the original code and license, see
# https://github.com/Qiskit/qiskit/blob/stable/1.0/qiskit/transpiler/passes/layout/apply_layout.py
#
# Modifications:
# - Renamed class to ApplyMitiqLayout
# - Removed layout length check
# - Added QuantumRegisters as parameter for class
# - Removed check for no post_layout
# - Improved typing
# - Modified docstring format
# - Modified length check

from typing import List

from qiskit import QuantumRegister
from qiskit.dagcircuit import DAGCircuit
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.exceptions import TranspilerError


class ApplyMitiqLayout(TransformationPass):  # type: ignore
    """Transform a circuit with virtual qubits into a circuit with physical
    qubits specified in the given list of QuantumRegisters.

    Transforms a DAGCircuit with virtual qubits into a DAGCircuit with physical
    qubits by applying the Layout given in `property_set`.
    Requires either of passes to set/select Layout, e.g. `SetLayout`,
    `TrivialLayout`. Assumes the Layout has full physical qubits.
    """

    def __init__(self, new_qregs: List[QuantumRegister]) -> None:
        """ApplyMitiqLayout constructor.

        Args:
            new_qregs: The new quantum registers for the circuit.
        """
        super().__init__()
        self._new_qregs = new_qregs

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        """Run the ApplyLayout pass on ``dag``.

        Args:
            dag (DAGCircuit): DAG to map.

        Returns:
            DAGCircuit: A mapped DAG (with physical qubits).

        Raises:
            TranspilerError: if no layout is found in ``property_set`` or
                             no full physical qubits.
        """
        layout = self.property_set["layout"]
        if not layout:
            raise TranspilerError(
                "No 'layout' is found in property_set. "
                "Please run a Layout pass in advance."
            )

        q = [qbit for qreg in self._new_qregs for qbit in qreg]
        if len(q) < len(dag.qubits):
            raise TranspilerError(
                "New layout has less qubits than DAGCircuit."
            )

        new_dag = DAGCircuit()

        for qreg in self._new_qregs:
            new_dag.add_qreg(qreg)

        new_dag.metadata = dag.metadata
        new_dag.add_clbits(dag.clbits)
        for creg in dag.cregs.values():
            new_dag.add_creg(creg)
        self.property_set["original_qubit_indices"] = {
            bit: index for index, bit in enumerate(dag.qubits)
        }
        for qreg in dag.qregs.values():
            self.property_set["layout"].add_register(qreg)
        virtual_physical_map = layout.get_virtual_bits()
        for node in dag.topological_op_nodes():
            qargs = [q[virtual_physical_map[qarg]] for qarg in node.qargs]
            new_dag.apply_operation_back(
                node.op, qargs, node.cargs, check=False
            )

        new_dag._global_phase = dag._global_phase

        return new_dag


class ClearLayout(TransformationPass):  # type: ignore
    """Clears the layout of the DAGCircuit"""

    def __init__(self) -> None:
        """ClearLayout"""
        super().__init__()

    def run(self, dag: DAGCircuit) -> None:
        """Clears the layout from the DAGCircuit's `property_set`."""
        self.property_set.pop("layout", None)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
"""Functions to calculate parameters for depolarizing noise and biased noise
models via a learning-based technique."""

from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
from scipy.optimize import minimize

from mitiq import QPROGRAM, Executor, Observable
from mitiq.cdr import generate_training_circuits
from mitiq.pec import execute_with_pec
from mitiq.pec.representations.biased_noise import (
    represent_operation_with_local_biased_noise,
)
from mitiq.pec.representations.depolarizing import (
    represent_operation_with_local_depolarizing_noise,
)


def learn_biased_noise_parameters(
    operations_to_learn: List[QPROGRAM],
    circuit: QPROGRAM,
    ideal_executor: Executor,
    noisy_executor: Executor,
    pec_kwargs: Dict[str, Any] = {},
    num_training_circuits: int = 5,
    fraction_non_clifford: float = 0.2,
    training_random_state: Optional[np.random.RandomState] = None,
    epsilon0: float = 0.05,
    eta0: float = 1,
    observable: Optional[Observable] = None,
    **learning_kwargs: Dict["str", Any],
) -> Tuple[bool, float, float]:
    r"""This function learns the biased noise parameters epsilon and eta
    associated to a set of input operations. The learning process is based on
    the execution of a set of training circuits on a noisy backend and on a
    classical simulator. The training circuits are near-Clifford approximations
    of the input circuit. A biased noise model characterization is assumed.

    Args:
        operations_to_learn: The ideal operations to learn the noise model of.
        circuit: The full quantum program as defined by the user.
        ideal_executor: Simulates a circuit and returns
            a noiseless ``QuantumResult``.
        noisy_executor: Executes a circuit on a noisy backend
            and returns a ``QuantumResult``.
        pec_kwargs (optional): Options to pass to ``execute_w_pec`` for the
            error-mitigated expectation value obtained from executing
            the training circuits.
        num_training_circuits: Number of near-Clifford circuits to be
            generated for training.
        fraction_non_clifford: The (approximate) fraction of non-Clifford
            gates in each training circuit.
        training_random_state: Seed for sampling the training circuits.
        epsilon0: Initial guess for noise strength.
        eta0: Initial guess for noise bias.
        observable (optional): Observable to compute the expectation value of.
            If None, the ``executor`` must return an expectation value.
            Otherwise the `QuantumResult` returned by `executor` is used to
            compute the expectation of the observable.
        learning_kwargs (optional): Additional data and options including
            ``method`` an optimization method supported by
            ``scipy.optimize.minimize`` and settings for the chosen
            optimization method.

    Returns:
        A 3-tuple containing a flag indicating whether or not the optimizer
        exited successfully, the optimized noise strength epsilon, and the
        optimized noise bias, eta.

    .. note:: Using this function may require some tuning. One of the main
        challenges is setting a good value of ``num_samples`` in the PEC
        options ``pec_kwargs``. Setting a small value of ``num_samples`` is
        typically necessary to obtain a reasonable execution time. On the other
        hand, using a number of PEC samples that is too small can result in a
        large statistical error, ultimately causing the optimization process to
        fail.
    """
    training_circuits = generate_training_circuits(
        circuit=circuit,
        num_training_circuits=num_training_circuits,
        fraction_non_clifford=fraction_non_clifford,
        random_state=training_random_state,
    )

    ideal_values = np.array(
        ideal_executor.evaluate(training_circuits, observable)
    )

    pec_data, method, minimize_kwargs = _parse_learning_kwargs(
        learning_kwargs=learning_kwargs
    )

    result = minimize(
        biased_noise_loss_function,
        [epsilon0, eta0],
        args=(
            operations_to_learn,
            training_circuits,
            ideal_values,
            noisy_executor,
            pec_kwargs,
            pec_data,
            observable,
        ),
        method=method,
        **minimize_kwargs,
    )

    success = result.success
    epsilon_opt, eta_opt = result.x

    return success, epsilon_opt, eta_opt


def learn_depolarizing_noise_parameter(
    operations_to_learn: List[QPROGRAM],
    circuit: QPROGRAM,
    ideal_executor: Executor,
    noisy_executor: Executor,
    pec_kwargs: Dict[str, Any] = {},
    num_training_circuits: int = 5,
    fraction_non_clifford: float = 0.2,
    training_random_state: Optional[np.random.RandomState] = None,
    epsilon0: float = 0.05,
    observable: Optional[Observable] = None,
    **learning_kwargs: Dict["str", Any],
) -> Tuple[bool, float]:
    r"""This function learns the depolarizing noise parameter (epsilon)
    associated to a set of input operations. The learning process is based on
    the execution of a set of training circuits on a noisy backend and on a
    classical simulator. The training circuits are near-Clifford approximations
    of the input circuit. A depolarizing noise model characterization is
    assumed.

    Args:
        operations_to_learn: The ideal operations to learn the noise model of.
        circuit: The full quantum program as defined by the user.
        ideal_executor: Simulates a circuit and returns
            a noiseless ``QuantumResult``.
        noisy_executor: Executes a circuit on a noisy backend
            and returns a ``QuantumResult``.
        pec_kwargs (optional): Options to pass to ``execute_w_pec`` for the
            error-mitigated expectation value obtained from executing
            the training circuits.
        num_training_circuits: Number of near-Clifford circuits to be
            generated for training.
        fraction_non_clifford: The (approximate) fraction of non-Clifford
            gates in each training circuit.
        training_random_state: Seed for sampling the training circuits.
        epsilon0: Initial guess for noise strength.
        observable (optional): Observable to compute the expectation value of.
            If None, the ``executor`` must return an expectation value.
            Otherwise the `QuantumResult` returned by `executor` is used to
            compute the expectation of the observable.
        learning_kwargs (optional): Additional data and options including
            ``method`` an optimization method supported by
            ``scipy.optimize.minimize`` and settings for the chosen
            optimization method.

    Returns:
        A 2-tuple containing flag indicating whether or not the optimizer
        exited successfully and the optimized noise strength epsilon.

    .. note:: Using this function may require some tuning. One of the main
        challenges is setting a good value of ``num_samples`` in the PEC
        options ``pec_kwargs``. Setting a small value of ``num_samples`` is
        typically necessary to obtain a reasonable execution time. On the other
        hand, using a number of PEC samples that is too small can result in a
        large statistical error, ultimately causing the optimization process to
        fail.
    """
    training_circuits = generate_training_circuits(
        circuit=circuit,
        num_training_circuits=num_training_circuits,
        fraction_non_clifford=fraction_non_clifford,
        random_state=training_random_state,
    )

    ideal_values = np.array(
        ideal_executor.evaluate(training_circuits, observable)
    )

    pec_data, method, minimize_kwargs = _parse_learning_kwargs(
        learning_kwargs=learning_kwargs
    )

    result = minimize(
        depolarizing_noise_loss_function,
        epsilon0,
        args=(
            operations_to_learn,
            training_circuits,
            ideal_values,
            noisy_executor,
            pec_kwargs,
            pec_data,
            observable,
        ),
        method=method,
        **minimize_kwargs,
    )

    success = result.success
    epsilon_opt = result.x[0]

    return success, epsilon_opt


def depolarizing_noise_loss_function(
    epsilon: npt.NDArray[np.float64],
    operations_to_mitigate: List[QPROGRAM],
    training_circuits: List[QPROGRAM],
    ideal_values: npt.NDArray[np.float64],
    noisy_executor: Executor,
    pec_kwargs: Dict[str, Any],
    pec_data: Optional[npt.NDArray[np.float64]] = None,
    observable: Optional[Observable] = None,
) -> float:
    r"""Loss function for optimizing quasi-probability representations
    assuming a depolarizing noise model depending on one real parameter.

    Args:
        epsilon: Array of optimization parameters epsilon
            (local noise strength) and eta (noise bias between reduced
            dephasing and depolarizing channels).
        operations_to_mitigate: List of ideal operations to be represented by
            a combination of noisy operations.
        training_circuits: List of training circuits for generating the
            error-mitigated expectation values.
        ideal_values: Expectation values obtained by noiseless simulations.
        noisy_executor: Executes the circuit with noise and returns a
            `QuantumResult`.
        pec_kwargs: Options to pass to `execute_w_pec` for the error-mitigated
            expectation value obtained from executing the training circuits.
        pec_data (optional): 2-D array of error-mitigated expection values for
            model training.
        observable (optional): Observable to compute the expectation value of.
            If None, the ``executor`` must return an expectation value.
            Otherwise the ``QuantumResult`` returned by ``executor`` is used to
            compute the expectation of the observable.

    Returns: Mean squared error between the error-mitigated values and
        the ideal values, over the training set.
    """
    if pec_data is not None:
        ind = np.abs(pec_data[:, 0] - epsilon).argmin()
        mitigated_values = pec_data[ind, 1:]

    else:
        representations = [
            represent_operation_with_local_depolarizing_noise(
                operation,
                epsilon[0],
            )
            for operation in operations_to_mitigate
        ]
        mitigated_values = np.array(
            [
                execute_with_pec(
                    circuit=training_circuit,
                    observable=observable,
                    executor=noisy_executor,
                    representations=representations,
                    full_output=False,
                    **pec_kwargs,
                )
                for training_circuit in training_circuits
            ]
        )

    return np.mean((mitigated_values - ideal_values) ** 2)


def biased_noise_loss_function(
    params: npt.NDArray[np.float64],
    operations_to_mitigate: List[QPROGRAM],
    training_circuits: List[QPROGRAM],
    ideal_values: npt.NDArray[np.float64],
    noisy_executor: Executor,
    pec_kwargs: Dict[str, Any],
    pec_data: Optional[npt.NDArray[np.float64]] = None,
    observable: Optional[Observable] = None,
) -> float:
    r"""Loss function for optimizing quasi-probability representations
    assuming a biased noise model depending on two real parameters.

    Args:
        params: Array of optimization parameters epsilon
            (local noise strength) and eta (noise bias between reduced
            dephasing and depolarizing channels).
        operations_to_mitigate: List of ideal operations to be represented by
            a combination of noisy operations.
        training_circuits: List of training circuits for generating the
            error-mitigated expectation values.
        ideal_values: Expectation values obtained by noiseless simulations.
        noisy_executor: Executes the circuit with noise and returns a
            `QuantumResult`.
        pec_kwargs: Options to pass to `execute_w_pec` for the error-mitigated
            expectation value obtained from executing the training circuits.
        pec_data (optional): 3-D array of error-mitigated expection values for
            model training.
        observable (optional): Observable to compute the expectation value of.
            If None, the ``executor`` must return an expectation value.
            Otherwise the ``QuantumResult`` returned by ``executor`` is used to
            compute the expectation of the observable.

    Returns: Mean squared error between the error-mitigated values and
        the ideal values, over the training set.
    """
    epsilon = params[0]
    eta = params[1]

    if pec_data is not None:
        ind_eps = np.abs(pec_data[:, 0, 0] - epsilon).argmin()
        ind_eta = np.abs(pec_data[0, :, 0] - eta).argmin()
        mitigated_values = pec_data[ind_eps, ind_eta, :]

    else:
        representations = [
            represent_operation_with_local_biased_noise(
                operation,
                epsilon,
                eta,
            )
            for operation in operations_to_mitigate
        ]
        mitigated_values = np.array(
            [
                execute_with_pec(
                    circuit=training_circuit,
                    observable=observable,
                    executor=noisy_executor,
                    representations=representations,
                    full_output=False,
                    **pec_kwargs,
                )
                for training_circuit in training_circuits
            ]
        )

    return np.mean((mitigated_values - ideal_values) ** 2)


def _parse_learning_kwargs(
    learning_kwargs: Dict[str, Any],
) -> Tuple[npt.NDArray[np.float64], str, Dict[str, Any]]:
    r"""Function for handling additional options and data for the learning
    functions.

    Args:
        learning_kwargs: Additional data and options including
            ``pec_data`` from pre-executed runs of PEC on training circuits,
            ``method`` an optimization method supported by
            ``scipy.optimize.minimize``, and settings for the chosen
            optimization method.

    Returns:
       Values contained in learning_kwargs or defaults if not specified.
    """
    minimize_kwargs = learning_kwargs.get("learning_kwargs")

    if minimize_kwargs is None:
        minimize_kwargs = {}

    pec_data = minimize_kwargs.pop("pec_data", None)
    method = minimize_kwargs.pop("method", "Nelder-Mead")

    return pec_data, method, minimize_kwargs



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

from mitiq.pec.types import NoisyOperation, OperationRepresentation, NoisyBasis
from mitiq.pec.sampling import sample_sequence, sample_circuit
from mitiq.pec.pec import execute_with_pec, mitigate_executor, pec_decorator

from mitiq.pec.representations import (
    represent_operation_with_global_depolarizing_noise,
    represent_operation_with_local_depolarizing_noise,
    represent_operations_in_circuit_with_global_depolarizing_noise,
    represent_operations_in_circuit_with_local_depolarizing_noise,
)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

# Quantum computer input/output.
from mitiq.typing import (
    Bitstring,
    MeasurementResult,
    QPROGRAM,
    QuantumResult,
    SUPPORTED_PROGRAM_TYPES,
)

# Utils
from mitiq.utils import qem_methods

# Executors and observables.
from mitiq.executor import Executor
from mitiq.observable import PauliString, Observable

# About and version.
from mitiq._about import about
from mitiq._version import __version__

# Calibration
from mitiq.calibration import (
    Calibrator,
    execute_with_mitigation,
    ZNE_SETTINGS,
    Settings,
)



# Copyright (C) Unitary Fund
# Portions of this code have been adapted from PennyLane's tutorial
# on Classical Shadows.
# Original authors: PennyLane developers: Brian Doolittle, Roeland Wiersema
# Tutorial link: https://pennylane.ai/qml/demos/tutorial_classical_shadows
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Defines utility functions for classical shadows protocol."""

from typing import Generator, List, Optional, Tuple

import numpy as np
import numpy.typing as npt
from scipy.linalg import sqrtm

import mitiq


def create_string(str_len: int, loc_list: List[int]) -> str:
    """
    This function returns a string of length ``str_len`` with 1s at the
    locations specified by ``loc_list`` and 0s elsewhere.

    Args:
        str_len: The length of the string.
        loc_list: A list of integers indices specifying the locations of 1s in
            the string.
    Returns:
        A bitstring constructed as above.

    Example:
        A basic example::

            create_string(5, [1, 3])
            >>> "01010"
    """
    return "".join(
        map(lambda i: "1" if i in set(loc_list) else "0", range(str_len))
    )


def valid_bitstrings(
    num_qubits: int, max_hamming_weight: Optional[int] = None
) -> set[str]:
    """
    Description.

    Args:
        num_qubits:
        max_hamming_weight:

    Returns:
        The set of all valid bitstrings on ``num_qubits`` bits, with a maximum
        hamming weight.
    Raises:
        Value error when ``max_hamming_weight`` is not greater than 0.
    """
    if max_hamming_weight and max_hamming_weight < 1:
        raise ValueError(
            "max_hamming_weight must be greater than 0. "
            f"Got {max_hamming_weight}."
        )

    bitstrings = {
        bin(i)[2:].zfill(num_qubits)
        for i in range(2**num_qubits)
        if bin(i).count("1") <= (max_hamming_weight or num_qubits)
    }
    return bitstrings


def fidelity(
    sigma: npt.NDArray[np.complex64], rho: npt.NDArray[np.complex64]
) -> float:
    """
    Calculate the fidelity between two states.

    Args:
        sigma: A state in terms of square matrix or vector.
        rho: A state in terms square matrix or vector.

    Returns:
        Scalar corresponding to the fidelity.
    """
    if sigma.ndim == 1 and rho.ndim == 1:
        val = np.abs(np.dot(sigma.conj(), rho)) ** 2.0
    elif sigma.ndim == 1 and rho.ndim == 2:
        val = np.abs(sigma.conj().T @ rho @ sigma)
    elif sigma.ndim == 2 and rho.ndim == 1:
        val = np.abs(rho.conj().T @ sigma @ rho)
    elif sigma.ndim == 2 and rho.ndim == 2:
        val = np.abs(np.trace(sqrtm(sigma) @ rho @ sqrtm(sigma)))
    else:
        raise ValueError("Invalid input dimensions")
    return float(val)


def batch_calibration_data(
    data: Tuple[List[str], List[str]], num_batches: int
) -> Generator[Tuple[List[str], List[str]], None, None]:
    """Batch calibration into chunks of size batch_size.

    Args:
        data: The random Pauli measurement outcomes.
        batch_size: Size of each batch that will be processed.

    Yields:
        Tuples of bit strings and pauli strings.
    """
    bits, paulis = data
    batch_size = len(bits) // num_batches
    for i in range(0, len(bits), batch_size):
        yield bits[i : i + batch_size], paulis[i : i + batch_size]


def n_measurements_tomography_bound(epsilon: float, num_qubits: int) -> int:
    """
    This function returns the minimum number of classical shadows required
    for state reconstruction for achieving the desired accuracy.

    Args:
        epsilon: The error on the estimator.
        num_qubits: The number of qubits in the system.

    Returns:
        An integer that gives the number of snapshots required to satisfy the
        shadow bound.
    """
    return int(34 * (4**num_qubits) * epsilon ** (-2))


def local_clifford_shadow_norm(obs: mitiq.PauliString) -> float:
    """
    Calculate shadow norm of an operator with random unitary sampled from local
    Clifford group.

    Args:
        obs: A self-adjoint operator, i.e. mitiq.PauliString with real coffe.
    Returns:
        Shadow norm when unitary ensemble is local Clifford group.
    """
    opt = obs.matrix()
    norm = (
        np.linalg.norm(
            opt - np.trace(opt) / 2 ** int(np.log2(opt.shape[0])),
            ord=np.inf,
        )
        ** 2
    )
    return float(norm)


def n_measurements_opts_expectation_bound(
    error: float,
    observables: List[mitiq.PauliString],
    failure_rate: float,
) -> Tuple[int, int]:
    """
    This function returns the minimum number of classical shadows required and
    the number of groups "k" into which we need to split the shadows for
    achieving the desired accuracy and failure rate in operator expectation
    value estimation.

    Args:
        error: The error on the estimator.
        observables: List of mitiq.PauliString corresponding to the
            observables we intend to measure.
        failure_rate: Rate of failure for the bound to hold.

    Returns:
        Integers quantifying the number of snapshots required to satisfy
        the shadow bound and the chunk size required to attain the specified
        failure rate.
    """
    M = len(observables)
    K = 2 * np.log(2 * M / failure_rate)

    N = 34 * max(local_clifford_shadow_norm(o) for o in observables) / error**2
    return int(np.ceil(N * K)), int(K)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Functions to create a Mirror Quantum Volume Benchmarking circuit
as defined in https://arxiv.org/abs/2303.02108."""

from typing import Optional, cast

import cirq

from mitiq import QPROGRAM
from mitiq.benchmarks.quantum_volume_circuits import (
    generate_quantum_volume_circuit,
)
from mitiq.interface.conversions import convert_from_mitiq


def generate_mirror_qv_circuit(
    num_qubits: int,
    depth: int,
    decompose: bool = False,
    seed: Optional[int] = None,
    return_type: Optional[str] = None,
) -> QPROGRAM:
    """Generate a mirror quantum volume circuit with the given number of qubits
    and depth as defined in :cite:`Amico_2023_arxiv`.

    The generated circuit consists of a quantum volume circuit up to `depth/2`
    layers followed by an inverse of the quantum volume portion up to `depth/2`
    when `depth` is an even number.

    When `depth` is odd, the layers will be changed to `depth+1`.


    The ideal output bit-string is a string of zeroes.

    Args:
        num_qubits: The number of qubits in the generated circuit.
        depth: The number of layers in the generated circuit.
        decompose: Recursively decomposes the randomly sampled (numerical)
            unitary matrix gates into simpler gates.
        seed: Seed for generating random circuit.
        return_type: String which specifies the type of the returned
            circuits. See the keys of ``mitiq.SUPPORTED_PROGRAM_TYPES``
            for options. If ``None``, the returned circuits have type
            ``cirq.Circuit``.

    Returns:
        A quantum volume circuit acting on ``num_qubits`` qubits.
    """
    if depth <= 0:
        raise ValueError(
            "{} is invalid for the generated circuit depth.", depth
        )

    if depth % 2 == 0:
        first_half_depth = int(depth / 2)
    else:
        first_half_depth = int((depth + 1) / 2)

    qv_circuit, _ = generate_quantum_volume_circuit(
        num_qubits, first_half_depth, seed=seed, decompose=decompose
    )
    qv_circuit = cast(cirq.Circuit, qv_circuit)

    mirror_qv_circuit = qv_circuit + cirq.inverse(qv_circuit)

    if decompose:
        # Decompose random unitary gates into simpler gates.
        mirror_qv_circuit = cirq.Circuit(cirq.decompose(mirror_qv_circuit))

    return_type = "cirq" if not return_type else return_type
    return convert_from_mitiq(mirror_qv_circuit, return_type)



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.
"""Functions for scaling supported circuits by inserting layers of identity
gates."""

import random
from typing import Tuple

import numpy as np
from cirq import Circuit, Moment, ops

from mitiq.interface import accept_qprogram_and_validate
from mitiq.utils import _append_measurements, _pop_measurements


class UnscalableCircuitError(Exception):
    pass


# Helper functions for identity scaling
def _check_scalable(input_circuit: Circuit) -> None:
    """Raises an error if the input circuit cannot be scaled by inserting
    identity layers.

    Args:
        circuit: Checks whether this circuit can be scaled.

    Raises:
        UnscalableCircuitError:
            * If the circuit has non-terminal measurements.
    """
    if not input_circuit.are_all_measurements_terminal():
        raise UnscalableCircuitError(
            "Circuit contains intermediate measurements and cannot be folded."
        )


def _calculate_id_layers(
    input_circuit_depth: int, scale_factor: float
) -> Tuple[int, int]:
    """Returns a tuple of integers that describes the number of identity layers
    to be inserted after each layer of the input circuit.

    Args:
        input_circuit_depth : Number of moments in the input_circuit
        scale_factor : Intended noise scaling factor

    Returns:
        (num_uniform_layers, num_partial_layers) : A tuple of the number of
        uniform identity layers to be inserted after each moment in the
        input_circuit and a number of partial layers to be inserted after
        some random moments to be able to achieve the intended scale factor.
    """
    if scale_factor < 1:
        raise ValueError(
            f"Requires scale_factor >= 1 but scale_factor = {scale_factor}."
        )

    num_uniform_layers = int(scale_factor - 1)
    int_scale_factor = num_uniform_layers + 1
    if np.isclose(int_scale_factor, scale_factor):
        return (num_uniform_layers, 0)
    else:
        num_partial_layers = int(
            input_circuit_depth * (scale_factor - 1 - num_uniform_layers)
        )
        return (num_uniform_layers, num_partial_layers)


@accept_qprogram_and_validate
def insert_id_layers(input_circuit: Circuit, scale_factor: float) -> Circuit:
    """Returns a scaled version of the input circuit by inserting layers of
    identities.

    Args:
        input_circuit : Cirq Circuit to be scaled
        scale_factor : Noise scaling factor as a float

    Returns:
        scaled_circuit : Scaled quantum circuit via identity layer insertions
    """
    _check_scalable(input_circuit)
    measurements = _pop_measurements(input_circuit)
    input_circuit_depth = len(input_circuit)

    num_uniform_layers, num_partial_layers = _calculate_id_layers(
        input_circuit_depth, scale_factor
    )

    random_moment_indices = random.sample(
        range(input_circuit_depth), num_partial_layers
    )

    circuit_qubits = input_circuit.all_qubits()
    id_layer = Moment(ops.I.on_each(*circuit_qubits))

    scaled_circuit = Circuit()
    for i, moment in enumerate(input_circuit):
        scaled_circuit.append(moment)

        scaled_circuit.append([id_layer] * (num_uniform_layers))

        if i in random_moment_indices:
            scaled_circuit.append(id_layer)

    _append_measurements(scaled_circuit, measurements)
    return scaled_circuit



# Copyright (C) Unitary Fund
#
# This source code is licensed under the GPL license (v3) found in the
# LICENSE file in the root directory of this source tree.

"""Classes corresponding to different zero-noise extrapolation methods."""

import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from cirq import Circuit
from matplotlib.figure import Figure
from numpy.lib.polynomial import RankWarning
from scipy.optimize import OptimizeWarning, curve_fit

from mitiq import QPROGRAM, QuantumResult
from mitiq.executor import Executor
from mitiq.interface import accept_any_qprogram_as_input
from mitiq.observable import Observable
from mitiq.zne.scaling import fold_gates_at_random

ExtrapolationResult = Union[
    float,  # The zero-noise value.
    Tuple[
        float,  # The zero-noise value.
        Optional[float],  # The (estimated) error on the zero-noise value.
        List[float],  # Optimal parameters found during fitting.
        Optional[np.ndarray],  # Covariance of fitting parameters.
        Callable[[float], float],  # Function that was fit.
    ],
]


class ExtrapolationError(Exception):
    """Error raised by :class:`.Factory` objects when
    the extrapolation fit fails.
    """

    pass


_EXTR_ERR = (
    "The extrapolation fit failed to converge."
    " The problem may be solved by switching to a more stable"
    " extrapolation model such as `LinearFactory`."
)


class ExtrapolationWarning(Warning):
    """Warning raised by :class:`.Factory` objects when
    the extrapolation fit is ill-conditioned.
    """

    pass


_EXTR_WARN = (
    "The extrapolation fit may be ill-conditioned."
    " Likely, more data points are necessary to fit the parameters"
    " of the model."
)

DATA_MISSING_ERR = (
    "Data is either ill-defined or not enough to evaluate the required"
    " information. Please make sure that the 'run' and 'reduce' methods"
    " have been called and that enough expectation values have been measured."
)


class ConvergenceWarning(Warning):
    """Warning raised by :class:`.Factory` objects when
    their `run_classical` method fails to converge.
    """

    pass


@accept_any_qprogram_as_input
def _check_circuit_length(circuit: Circuit) -> None:
    """Raises a warning if the circuit is too short."""
    if len(list(circuit.all_operations())) < 5:
        warnings.warn(
            "The input circuit is very short. "
            "This may reduce the accuracy of noise scaling."
        )


def mitiq_curve_fit(
    ansatz: Callable[..., float],
    scale_factors: Sequence[float],
    exp_values: Sequence[float],
    init_params: Optional[List[float]] = None,
) -> Tuple[List[float], npt.NDArray[np.float64]]:
    """Fits the ansatz to the (scale factor, expectation value) data using
    ``scipy.optimize.curve_fit``, returning the optimal parameters and
    covariance matrix of the parameters.

    Args:
        ansatz: The model function used for zero-noise extrapolation. The first
            argument is the noise scale variable, the remaining arguments are
            the parameters to fit.
        scale_factors: The array of noise scale factors.
        exp_values: The array of expectation values.
        init_params: Initial guess for the parameters. If None, the initial
            values are set to 1.

    Returns:
        The array of optimal parameters and the covariance matrix of the
        parameters. If the fit is ill-conditioned, the covariance matrix may
        contain np.inf elements.

    Raises:
        ExtrapolationError: If the extrapolation fit fails.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """
    try:
        with warnings.catch_warnings(record=True) as warn_list:
            opt_params, params_cov = curve_fit(
                ansatz, scale_factors, exp_values, p0=init_params
            )
        for warn in warn_list:
            # replace OptimizeWarning with ExtrapolationWarning
            if warn.category is OptimizeWarning:
                warn.category = ExtrapolationWarning
                warn.message = _EXTR_WARN
            # re-raise all warnings
            warnings.warn_explicit(
                warn.message, warn.category, warn.filename, warn.lineno
            )
    except RuntimeError:
        raise ExtrapolationError(_EXTR_ERR) from None
    return list(opt_params), params_cov


def mitiq_polyfit(
    scale_factors: Sequence[float],
    exp_values: Sequence[float],
    deg: int,
    weights: Optional[Sequence[float]] = None,
) -> Tuple[List[float], Optional[npt.NDArray[np.float64]]]:
    """Fits the ansatz to the (scale factor, expectation value) data using
    ``numpy.polyfit``, returning the optimal parameters and covariance matrix
    of the parameters.

    Args:
        scale_factors: The array of noise scale factors.
        exp_values: The array of expectation values.
        deg: The degree of the polynomial fit.
        weights: Optional array of weights for each sampled point.
            This is used to make a weighted least squares fit.

    Returns:
        The optimal parameters and covariance matrix of the parameters.
        If there is not enough data to estimate the covariance matrix, it is
        returned as None.

    Raises:
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """

    with warnings.catch_warnings(record=True) as warn_list:
        try:
            opt_params, params_cov = np.polyfit(
                scale_factors, exp_values, deg, w=weights, cov=True
            )
        except (ValueError, np.linalg.LinAlgError):
            opt_params = np.polyfit(scale_factors, exp_values, deg, w=weights)
            params_cov = None

    for warn in warn_list:
        # replace RankWarning with ExtrapolationWarning
        if warn.category is RankWarning:
            warn.category = ExtrapolationWarning
            warn.message = _EXTR_WARN
        # re-raise all warnings
        warnings.warn_explicit(
            warn.message, warn.category, warn.filename, warn.lineno
        )
    return list(opt_params), params_cov


class Factory(ABC):
    """Abstract base class which performs the classical parts of zero-noise
    extrapolation. This minimally includes:

        * scaling circuits,
        * sending jobs to execute,
        * collecting the results,
        * fitting the collected data,
        * Extrapolating to the zero-noise limit.

    If all scale factors are set a priori, the jobs can be batched. This is
    handled by a BatchedFactory.

    If the next scale factor depends on the previous history of results,
    jobs are run sequentially. This is handled by an AdaptiveFactory.
    """

    def __init__(self) -> None:
        self._instack: List[Dict[str, float]] = []
        self._outstack: List[float] = []
        self._opt_params: Optional[List[float]] = None
        self._params_cov: Optional[npt.NDArray[np.float64]] = None
        self._zne_limit: Optional[float] = None
        self._zne_error: Optional[float] = None
        self._zne_curve: Optional[Callable[[float], float]] = None
        self._already_reduced = False
        self._options: Dict[str, Optional[float]] = {}

    def get_scale_factors(self) -> List[float]:
        """Returns the scale factors at which the factory has computed
        expectation values.
        """
        return [params.get("scale_factor", 0.0) for params in self._instack]

    def get_expectation_values(self) -> List[float]:
        """Returns the expectation values computed by the factory."""
        return self._outstack

    def get_optimal_parameters(self) -> List[float]:
        """Returns the optimal model parameters produced by the extrapolation
        fit.
        """
        if self._opt_params is None:
            raise ValueError(DATA_MISSING_ERR)
        return self._opt_params

    def get_parameters_covariance(self) -> npt.NDArray[np.float64]:
        """Returns the covariance matrix of the model parameters produced by
        the extrapolation fit.
        """
        if self._params_cov is None:
            raise ValueError(DATA_MISSING_ERR)
        return self._params_cov

    def get_zero_noise_limit(self) -> float:
        """Returns the last evaluation of the zero-noise limit
        computed by the factory. To re-evaluate
        its value, the method 'reduce' should be called first.
        """
        if self._zne_limit is None:
            raise ValueError(DATA_MISSING_ERR)
        return self._zne_limit

    def get_zero_noise_limit_error(self) -> float:
        """Returns the extrapolation error representing the uncertainty
        affecting the zero-noise limit. It is deduced by error propagation
        from the covariance matrix associated to the fit parameters.

        Note: this quantity is only related to the ability of the model
            to fit the measured data. Therefore, it may underestimate the
            actual error existing between the zero-noise limit and the
            true ideal expectation value.
        """
        if self._zne_error is None:
            raise ValueError(DATA_MISSING_ERR)
        return self._zne_error

    def get_extrapolation_curve(self) -> Callable[[float], float]:
        """Returns the extrapolation curve, i.e., a function which
        inputs a noise scale factor and outputs the associated expectation
        value. This function is the solution of the regression problem
        used to evaluate the zero-noise extrapolation.
        """
        if self._zne_curve is None:
            raise ValueError(DATA_MISSING_ERR)
        return self._zne_curve

    @abstractmethod
    def run(
        self,
        qp: QPROGRAM,
        executor: Union[Executor, Callable[..., QuantumResult]],
        observable: Optional[Observable] = None,
        scale_noise: Callable[
            [QPROGRAM, float], QPROGRAM
        ] = fold_gates_at_random,  # type: ignore [has-type]
        num_to_average: int = 1,
    ) -> "Factory":
        """Calls the executor function on noise-scaled quantum circuit and
        stores the results.

        Args:
            qp: Quantum circuit to scale noise in.
            executor: A ``mitiq.Executor`` or a function which inputs a (list
                of) quantum circuits and outputs a (list of)
                ``mitiq.QuantumResult`` s.
            observable: Observable to compute the expectation value of. If
                None, the `executor` must return an expectation value.
                Otherwise, the `QuantumResult` returned by `executor` is used
                to compute the expectation of the observable.
            scale_noise: Function which inputs a quantum circuit and outputs
                a noise-scaled quantum circuit.
            num_to_average: Number of times the executor function is called
                on each noise-scaled quantum circuit.
        """
        raise NotImplementedError

    @abstractmethod
    def run_classical(
        self,
        scale_factor_to_expectation_value: Callable[..., float],
    ) -> "Factory":
        """Calls the function scale_factor_to_expectation_value at each scale
        factor of the factory, and stores the results.

        Args:
            scale_factor_to_expectation_value: A function which inputs a scale
                factor and outputs an expectation value. This does not have to
                involve a quantum processor making this a "classical analogue"
                of the run method.
        """
        raise NotImplementedError

    @abstractmethod
    def reduce(self) -> float:
        raise NotImplementedError

    def push(
        self, instack_val: Dict[str, float], outstack_val: float
    ) -> "Factory":
        """Appends "instack_val" to "self._instack" and "outstack_val" to
        "self._outstack". Each time a new expectation value is computed this
        method should be used to update the internal state of the Factory.
        """
        if self._already_reduced:
            warnings.warn(
                "You are pushing new data into a factory object despite its "
                ".reduce() method has already been called. Please make "
                "sure your intention is to append new data to the stack of "
                "previous data. Otherwise, the method .reset() can be used "
                "to clean the internal state of the factory.",
                ExtrapolationWarning,
            )
        self._instack.append(instack_val)
        self._outstack.append(outstack_val)
        return self

    def plot_data(self) -> Figure:
        """Returns a figure which is a scatter plot of (x, y) data where x are
        scale factors at which expectation values have been computed, and y are
        the associated expectation values.

        Returns:
            fig: A 2D scatter plot described above.
        """
        fig = plt.figure(figsize=(7, 5))
        ax = plt.gca()
        plt.plot(
            self.get_scale_factors(),
            self.get_expectation_values(),
            "o",
            markersize=10,
            markeredgecolor="black",
            alpha=0.8,
            label="Data",
        )
        ax.grid(True)
        plt.xlabel("Noise scale factor")
        plt.ylabel("Expectation value")
        return fig

    def plot_fit(self) -> Figure:
        """Returns a figure which plots the experimental data as well as the
        best fit curve.

        Returns:
            fig: A figure which plots the best fit curve as well as the data.
        """
        fig = self.plot_data()

        smooth_scale_factors = np.linspace(0, self.get_scale_factors()[-1], 20)
        smooth_expectations = [
            self.get_extrapolation_curve()(scale_factor)
            for scale_factor in smooth_scale_factors
        ]
        plt.xlim(left=0)
        fig.axes[0].plot(
            smooth_scale_factors,
            smooth_expectations,
            "--",
            lw=2,
            color="black",
            label="Best fit",
        )

        return fig

    def reset(self) -> "Factory":
        """Resets the internal state of the Factory."""

        self._instack = []
        self._outstack = []
        self._opt_params = None
        self._params_cov = None
        self._zne_limit = None
        self._zne_error = None
        self._already_reduced = False
        return self


class BatchedFactory(Factory, ABC):
    """Abstract class of a non-adaptive Factory initialized with a
    pre-determined set of scale factors.

    Specific (non-adaptive) extrapolation algorithms are derived from this
    class by defining the `reduce` method.
    """

    def __init__(
        self,
        scale_factors: Sequence[float],
        shot_list: Optional[List[int]] = None,
    ) -> None:
        """Constructs a BatchedFactory.

        Args:
            scale_factors: Sequence of noise scale factors at which expectation
                values should be measured.
            shot_list: Optional sequence of integers corresponding to the
                number of samples taken for each expectation value. If this
                argument is explicitly passed to the factory, it must have the
                same length of scale_factors and the executor function must
                accept "shots" as a valid keyword argument.

        Raises:
            ValueError: If the number of scale factors is less than 2.
            TypeError: If shot_list is provided and has any non-integer values.
        """
        if len(scale_factors) < 2:
            raise ValueError("At least 2 scale factors are necessary.")

        if shot_list and (
            not isinstance(shot_list, Sequence)
            or not all([isinstance(shots, int) for shots in shot_list])
        ):
            raise TypeError(
                "The optional argument shot_list must be None "
                "or a valid iterator of integers."
            )
        if shot_list and (len(scale_factors) != len(shot_list)):
            raise IndexError(
                "The arguments scale_factors and shot_list"
                " must have the same length."
                f" But len(scale_factors) is {len(scale_factors)}"
                f" and len(shot_list) is {len(shot_list)}."
            )

        self._scale_factors = scale_factors
        self._shot_list = shot_list

        super(BatchedFactory, self).__init__()

    @staticmethod
    @abstractmethod
    def extrapolate(*args: Any, **kwargs: Any) -> ExtrapolationResult:
        """Returns the extrapolation to the zero-noise limit."""
        raise NotImplementedError

    def reduce(self) -> float:
        """Evaluates the zero-noise limit found by fitting according to
        the factory's extrapolation method.

        Returns:
            The zero-noise limit.
        """
        (
            self._zne_limit,
            self._zne_error,
            self._opt_params,
            self._params_cov,
            self._zne_curve,
        ) = self.extrapolate(  # type: ignore
            self.get_scale_factors(),
            self.get_expectation_values(),
            full_output=True,
            **self._options,
        )

        self._already_reduced = True
        return self._zne_limit

    def run(
        self,
        qp: QPROGRAM,
        executor: Union[Executor, Callable[..., QuantumResult]],
        observable: Optional[Observable] = None,
        scale_noise: Callable[
            [QPROGRAM, float], QPROGRAM
        ] = fold_gates_at_random,  # type: ignore [has-type]
        num_to_average: int = 1,
    ) -> "BatchedFactory":
        """Computes the expectation values at each scale factor and stores them
        in the factory. If the executor returns a single expectation value, the
        circuits are run sequentially. If the executor is batched and returns
        a list of expectation values (one for each circuit), then the circuits
        are sent to the backend as a single job. To detect if an executor is
        batched, it must be annotated with a return type that is one of the
        following:

            * Iterable[float]
            * List[float]
            * Sequence[float]
            * Tuple[float]
            * numpy.ndarray

        Args:
            qp: Quantum circuit to run.
            executor: A ``mitiq.Executor`` or a function which inputs a (list
                of) quantum circuits and outputs a (list of)
                ``mitiq.QuantumResult`` s.
            observable: Observable to compute the expectation value of. If
                None, the `executor` must return an expectation value.
                Otherwise, the `QuantumResult` returned by `executor` is used
                to compute the expectation of the observable.
            scale_noise: Noise scaling function.
            num_to_average: The number of circuits executed for each noise
                scale factor. This parameter can be used to increase the
                precision of the "executor" or to average the effect of a
                non-deterministic "scale_noise" function.
        """
        self.reset()
        self._batch_populate_instack()

        _check_circuit_length(qp)

        # Get all noise-scaled circuits to run.
        to_run = self._generate_circuits(qp, scale_noise, num_to_average)

        # Run all circuits.
        if not isinstance(executor, Executor):
            executor = Executor(executor)

        # Get the list of keywords associated to each circuit in "to_run".
        kwargs_list = self._get_keyword_args(num_to_average)

        # If there are different keyword args, run each circuit individually.
        # https://stackoverflow.com/questions/1151658/python-hashable-dicts.
        class HashableDict(Dict[Any, Any]):
            def __hash__(self) -> int:  # type: ignore[override]
                return hash(tuple(sorted(self.items())))

        if len(set(HashableDict(kwargs) for kwargs in kwargs_list)) != 1:
            res = []
            for circuit, kwargs in zip(to_run, kwargs_list):
                res.extend(
                    executor.evaluate(
                        circuit, observable, force_run_all=True, **kwargs
                    )
                )
        else:
            # Else, run all circuits.
            res = executor.evaluate(
                to_run, observable, force_run_all=True, **kwargs_list[0]
            )

        # Reshape "res" to have "num_to_average" columns
        reshaped = np.array(res).reshape((-1, num_to_average))

        # Average the "num_to_average" columns
        self._outstack = np.average(reshaped, axis=1).tolist()

        return self

    def run_classical(
        self, scale_factor_to_expectation_value: Callable[..., float]
    ) -> "BatchedFactory":
        """Computes expectation values by calling the input function at each
        scale factor.

        Args:
            scale_factor_to_expectation_value: Function mapping a noise scale
                factor to an expectation value. If shot_list is not None,
                "shots" must be an argument of this function.
        """
        self.reset()
        self._batch_populate_instack()
        kwargs_list = self._get_keyword_args(num_to_average=1)

        self._outstack = [
            scale_factor_to_expectation_value(scale_factor, **kwargs)
            for scale_factor, kwargs in zip(self._scale_factors, kwargs_list)
        ]

        return self

    def _generate_circuits(
        self,
        circuit: QPROGRAM,
        scale_noise: Callable[[QPROGRAM, float], QPROGRAM],
        num_to_average: int = 1,
    ) -> List[QPROGRAM]:
        """Returns all noise-scaled circuits to run.

        Args:
            circuit: Base circuit to scale noise in.
            scale_noise: Noise scaling function.
            num_to_average: Number of times to call scale_noise at each scale
                factor.
        """
        to_run = []
        for scale_factor in self.get_scale_factors():
            for _ in range(num_to_average):
                to_run.append(scale_noise(circuit, scale_factor))
        return to_run

    def _batch_populate_instack(self) -> None:
        """Populates the instack with all computed values."""
        if self._shot_list:
            self._instack = [
                {"scale_factor": scale, "shots": shots}
                for scale, shots in zip(self._scale_factors, self._shot_list)
            ]
        else:
            self._instack = [
                {"scale_factor": scale} for scale in self._scale_factors
            ]

    def _get_keyword_args(self, num_to_average: int) -> List[Dict[str, Any]]:
        """Returns a list of keyword dictionaries to be used for
        executing the circuits generated by the method "_generate_circuits".

        Args:
            num_to_average: The number of times the same keywords are used
                for each scale factor. This should correspond to the number
                of circuits executed for each scale factor.

        Returns:
            The output list of keyword dictionaries.
        """
        params = deepcopy(self._instack)
        for d in params:
            _ = d.pop("scale_factor")

        # Repeat each keyword num_to_average times
        return [k for k in params for _ in range(num_to_average)]


class AdaptiveFactory(Factory, ABC):
    """Abstract class designed to adaptively produce a new noise scaling
    parameter based on a historical stack of previous noise scale parameters
    ("self._instack") and previously estimated expectation values
    ("self._outstack").

    Specific zero-noise extrapolation algorithms which are adaptive are derived
    from this class.
    """

    @abstractmethod
    def next(self) -> Dict[str, float]:
        """Returns a dictionary of parameters to execute a circuit at."""
        raise NotImplementedError

    @abstractmethod
    def is_converged(self) -> bool:
        """Returns True if all needed expectation values have been computed,
        else False.
        """
        raise NotImplementedError

    @abstractmethod
    def reduce(self) -> float:
        """Returns the extrapolation to the zero-noise limit."""
        raise NotImplementedError

    def run_classical(
        self,
        scale_factor_to_expectation_value: Callable[..., float],
        max_iterations: int = 100,
    ) -> "AdaptiveFactory":
        """Evaluates a sequence of expectation values until enough
        data is collected (or iterations reach "max_iterations").

        Args:
            scale_factor_to_expectation_value: Function mapping a noise scale
                factor to an expectation value. If shot_list is not None,
                "shots" must be an argument of this function.
            max_iterations: Maximum number of iterations (optional).
                Default: 100.

        Raises:
            ConvergenceWarning: If iteration loop stops before convergence.
        """
        # Reset the instack, outstack, and optimal parameters
        self.reset()

        counter = 0
        while not self.is_converged() and counter < max_iterations:
            next_in_params = self.next()
            next_exec_params = deepcopy(next_in_params)

            # Get next scale factor and remove it from next_exec_params
            scale_factor = next_exec_params.pop("scale_factor")
            next_expval = scale_factor_to_expectation_value(
                scale_factor, **next_exec_params
            )
            self.push(next_in_params, next_expval)
            counter += 1

        if counter == max_iterations:
            warnings.warn(
                "Factory iteration loop stopped before convergence. "
                f"Maximum number of iterations ({max_iterations}) "
                "was reached.",
                ConvergenceWarning,
            )

        return self

    def run(
        self,
        qp: QPROGRAM,
        executor: Union[Executor, Callable[..., QuantumResult]],
        observable: Optional[Observable] = None,
        scale_noise: Callable[
            [QPROGRAM, float], QPROGRAM
        ] = fold_gates_at_random,  # type: ignore [has-type]
        num_to_average: int = 1,
        max_iterations: int = 100,
    ) -> "AdaptiveFactory":
        """Evaluates a sequence of expectation values by executing quantum
        circuits until enough data is collected (or iterations reach
        "max_iterations").

        Args:
            qp: Circuit to mitigate.
            executor: A ``mitiq.Executor`` or a function which inputs a (list
                of) quantum circuits and outputs a (list of)
                ``mitiq.QuantumResult`` s.
            observable: Observable to compute the expectation value of. If
                None, the `executor` must return an expectation value.
                Otherwise, the `QuantumResult` returned by `executor` is used
                to compute the expectation of the observable.
            scale_noise: Function that scales the noise level of a quantum
                circuit.
            num_to_average: Number of times expectation values are computed by
                the executor after each call to scale_noise, then averaged.
            max_iterations: Maximum number of iterations (optional).
        """

        _check_circuit_length(qp)

        if not isinstance(executor, Executor):
            executor = Executor(executor)

        def scale_factor_to_expectation_value(
            scale_factor: float, **exec_params: Any
        ) -> float:
            """Evaluates the quantum expectation value for a given
            scale_factor and other executor parameters."""
            to_run = [
                scale_noise(qp, scale_factor) for _ in range(num_to_average)
            ]
            expectation_values = executor.evaluate(  # type: ignore[union-attr]
                to_run, observable, force_run_all=True, **exec_params
            )
            return cast(float, np.average(expectation_values))

        return self.run_classical(
            scale_factor_to_expectation_value, max_iterations
        )


class PolyFactory(BatchedFactory):
    """Factory object implementing a zero-noise extrapolation algorithm based
    on a polynomial fit.

    Args:
        scale_factors: Sequence of noise scale factors at which
            expectation values should be measured.
        order: Extrapolation order (degree of the polynomial fit).
            It cannot exceed len(scale_factors) - 1.
        shot_list: Optional sequence of integers corresponding to the number
            of samples taken for each expectation value. If this
            argument is explicitly passed to the factory, it must have
            the same length of scale_factors and the executor function
            must accept "shots" as a valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

    Note:
        RichardsonFactory and LinearFactory are special cases of PolyFactory.
    """

    def __init__(
        self,
        scale_factors: Sequence[float],
        order: int,
        shot_list: Optional[List[int]] = None,
    ) -> None:
        """Instantiates a new object of this Factory class."""
        if order > len(scale_factors) - 1:
            raise ValueError(
                "The extrapolation order cannot exceed len(scale_factors) - 1."
            )
        super(PolyFactory, self).__init__(scale_factors, shot_list)
        self._options = {"order": order}

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        order: int,
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates a polynomial extrapolation to the
        zero-noise limit.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            order: The extrapolation order (degree of the polynomial fit).
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional information about the
                extrapolated limit is returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """

        opt_params, params_cov = mitiq_polyfit(
            scale_factors, exp_values, order
        )

        zne_limit = opt_params[-1]

        if not full_output:
            return zne_limit

        zne_error = None

        if params_cov is not None:
            if params_cov.shape == (order + 1, order + 1):
                zne_error = np.sqrt(params_cov[order, order])

        def zne_curve(scale_factor: float) -> float:
            return cast(float, np.polyval(opt_params, scale_factor))

        return zne_limit, zne_error, opt_params, params_cov, zne_curve


class RichardsonFactory(BatchedFactory):
    """Factory object implementing Richardson extrapolation.

    Args:
        scale_factors: Sequence of noise scale factors at which
                       expectation values should be measured.
        shot_list: Optional sequence of integers corresponding to the number
                   of samples taken for each expectation value. If this
                   argument is explicitly passed to the factory, it must have
                   the same length of scale_factors and the executor function
                   must accept "shots" as a valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates the Richardson extrapolation to the
         zero-noise limit.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional results are returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """
        # Richardson extrapolation is a particular case of a polynomial fit
        # with order equal to the number of data points minus 1.
        order = len(scale_factors) - 1
        return PolyFactory.extrapolate(
            scale_factors, exp_values, order, full_output
        )


class FakeNodesFactory(BatchedFactory):
    """Factory object implementing a modified version [De2020polynomial]_ of
    Richardson extrapolation. In this version the original set of scale factors
    is mapped to a new set of fake nodes, known as Chebyshev-Lobatto points.
    This method may give a better interpolation for particular types of curves
    and if the number of scale factors is large (> 10). One should be aware
    that, in many other cases, the fake nodes extrapolation method is usually
    not superior to standard Richardson extrapolation.

    Args:
        scale_factors: Sequence of noise scale factors at which
                       expectation values should be measured.
        shot_list: Optional sequence of integers corresponding to the number
                   of samples taken for each expectation value. If this
                   argument is explicitly passed to the factory, it must have
                   the same length of scale_factors and the executor function
                   must accept "shots" as a valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

    .. [De2020polynomial] : S.De Marchia. F. Marchetti, E.Perracchionea
        and D.Poggialia,
        "Polynomial interpolation via mapped bases without resampling,"
        *Journ of Comp. and App. Math.* **364**, 112347 (2020),
        (https://www.sciencedirect.com/science/article/abs/pii/S0377042719303449).
    """

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        full_output: bool = False,
    ) -> ExtrapolationResult:
        if not FakeNodesFactory._is_equally_spaced(scale_factors):
            raise ValueError("The scale factors must be equally spaced.")

        # Define interval [a, b] for which the scale_factors are mapped to
        a = 0.0
        b = min(scale_factors) + max(scale_factors)

        # Mapping to the fake nodes
        fake_nodes = FakeNodesFactory._map_to_fake_nodes(scale_factors, a, b)

        if not full_output:
            return RichardsonFactory.extrapolate(fake_nodes, exp_values)

        (
            zne_limit,
            zne_error,
            opt_params,
            params_cov,
            zne_curve,
        ) = RichardsonFactory.extrapolate(  # type: ignore[misc]
            fake_nodes, exp_values, True
        )

        # Convert zne_curve from the "fake node space" to the real space.
        # Note: since a=0.0, this conversion is not necessary for zne_limit.
        def new_curve(scale_factor: float) -> float:
            """Get real zne_curve from the curve based on fake nodes."""
            return zne_curve(
                FakeNodesFactory._map_to_fake_nodes([scale_factor], a, b)[0]
            )

        return zne_limit, zne_error, opt_params, params_cov, new_curve

    @staticmethod
    def _map_to_fake_nodes(
        x: Sequence[float], a: float, b: float
    ) -> Sequence[float]:
        """
        A function that maps inputs to Chebyshev-Lobatto points. Based on
        the function [De2020polynomial]_:
            S(x) = (a - b)/2 * cos(pi * (x - a)/(b - a)) + (a + b)/2.
        Where a and b are the endpoints of the interval [a, b] of CL points
        we are mapping to.

        Args:
            x: Sequence[float]: Set of values to be mapped to CL points.
            a: A float representing the interval starting at a.
            b: A float representing the interval ending at b.
        Returns:
            A new sequence of fake nodes (Chebyshev-Lobatto points).

        .. [De2020polynomial]: S.De Marchia. F. Marchetti, E.Perracchionea
            and D.Poggialia,
            "Polynomial interpolation via mapped bases without resampling,"
            *Journ of Comp. and App. Math.* **364**, 112347 (2020),
            (https://www.sciencedirect.com/science/article/abs/pii/S0377042719303449).
        """

        # The mapping function
        def mapping(_x: float) -> float:
            return (a - b) / 2 * np.cos(np.pi * (_x - a) / (b - a)) + (
                a + b
            ) / 2

        return [mapping(y) for y in x]

    @staticmethod
    def _is_equally_spaced(arr: Sequence[float]) -> bool:
        """Checks if the sequence is equally spaced."""

        diff_arr = np.diff(np.sort(arr))
        return np.allclose(diff_arr, diff_arr[0])


class LinearFactory(BatchedFactory):
    """
    Factory object implementing zero-noise extrapolation based
    on a linear fit.

    Args:
        scale_factors: Sequence of noise scale factors at which
                       expectation values should be measured.
        shot_list: Optional sequence of integers corresponding to the number
                   of samples taken for each expectation value. If this
                   argument is explicitly passed to the factory, it must have
                   the same length of scale_factors and the executor function
                   must accept "shots" as a valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates the linear extrapolation to the
        zero-noise limit.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional results are returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """
        # Linear extrapolation is equivalent to a polynomial fit with order=1
        return PolyFactory.extrapolate(
            scale_factors, exp_values, 1, full_output
        )


class ExpFactory(BatchedFactory):
    """
    Factory object implementing a zero-noise extrapolation algorithm assuming
    an exponential ansatz y(x) = a + b * exp(-c * x), with c > 0.

    If y(x->inf) is unknown, the ansatz y(x) is fitted with a non-linear
    optimization.

    If y(x->inf) is given and avoid_log=False, the exponential
    model is mapped into a linear model by a logarithmic transformation.

    Args:
        scale_factors: Sequence of noise scale factors at which expectation
            values should be measured.
        asymptote: Infinite-noise limit (optional argument).
        avoid_log: If set to True, the exponential model is not linearized
            with a logarithm and a non-linear fit is applied even if asymptote
            is not None. The default value is False.
        shot_list: Optional sequence of integers corresponding to the number
            of samples taken for each expectation value. If this argument is
            explicitly passed to the factory, it must have the same length of
            scale_factors and the executor function must accept "shots" as a
            valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationError: If the extrapolation fit fails.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """

    def __init__(
        self,
        scale_factors: Sequence[float],
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        shot_list: Optional[List[int]] = None,
    ) -> None:
        """Instantiate an new object of this Factory class."""
        super(ExpFactory, self).__init__(scale_factors, shot_list)
        if not (asymptote is None or isinstance(asymptote, float)):
            raise ValueError(
                "The argument 'asymptote' must be either a float or None"
            )
        self._options = {
            "asymptote": asymptote,
            "avoid_log": avoid_log,
        }

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        eps: float = 1.0e-6,
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates the extrapolation to the zero-noise
        limit assuming an exponential ansatz y(x) = a + b * exp(-c * x),
        with c > 0.

        If y(x->inf) is unknown, the ansatz y(x) is fitted with a non-linear
        optimization.

        If y(x->inf) is given and avoid_log=False, the exponential
        model is mapped into a linear model by a logarithmic transformation.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            asymptote: The infinite-noise limit y(x->inf) (optional argument).
            avoid_log: If set to True, the exponential model is not linearized
                with a logarithm and a non-linear fit is applied even if
                asymptote is not None. The default value is False.
            eps: Epsilon to regularize log(sign(scale_factors - asymptote))
                when the argument is to close to zero or negative.
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional information about the
                extrapolated limit is returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ValueError: If the arguments are not consistent with the
                extrapolation model.
            ExtrapolationError: If the extrapolation fit fails.
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """
        return PolyExpFactory.extrapolate(
            scale_factors,
            exp_values,
            order=1,
            asymptote=asymptote,
            avoid_log=avoid_log,
            eps=eps,
            full_output=full_output,
        )


class PolyExpFactory(BatchedFactory):
    """
    Factory object implementing a zero-noise extrapolation algorithm assuming
    an (almost) exponential ansatz with a non linear exponent
    y(x) = a + sign * exp(z(x)), where z(x) is a polynomial of a given order.

    The parameter "sign" is a sign variable which can be either 1 or -1,
    corresponding to decreasing and increasing exponentials, respectively.
    The parameter "sign" is automatically deduced from the data.

    If y(x->inf) is unknown, the ansatz y(x) is fitted with a non-linear
    optimization.

    If y(x->inf) is given and avoid_log=False, the exponential
    model is mapped into a polynomial model by logarithmic transformation.

    Args:
        scale_factors: Sequence of noise scale factors at which
                       expectation values should be measured.
        order: Extrapolation order (degree of the polynomial z(x)).
               It cannot exceed len(scale_factors) - 1.
               If asymptote is None, order cannot exceed
               len(scale_factors) - 2.
        asymptote: The infinite-noise limit y(x->inf) (optional argument).
        avoid_log: If set to True, the exponential model is not linearized
                   with a logarithm and a non-linear fit is applied even
                   if asymptote is not None. The default value is False.
        shot_list: Optional sequence of integers corresponding to the number
                   of samples taken for each expectation value. If this
                   argument is explicitly passed to the factory, it must have
                   the same length of scale_factors and the executor function
                   must accept "shots" as a valid keyword argument.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationError: If the extrapolation fit fails.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.
    """

    def __init__(
        self,
        scale_factors: Sequence[float],
        order: int,
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        shot_list: Optional[List[int]] = None,
    ) -> None:
        """Instantiates a new object of this Factory class."""
        super(PolyExpFactory, self).__init__(scale_factors, shot_list)
        if not (asymptote is None or isinstance(asymptote, float)):
            raise ValueError(
                "The argument 'asymptote' must be either a float or None"
            )
        self._options = {
            "order": order,
            "asymptote": asymptote,
            "avoid_log": avoid_log,
        }

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        order: int,
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        eps: float = 1.0e-6,
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates the extrapolation to the
        zero-noise limit with an exponential ansatz (whose exponent
        is a polynomial of degree "order").

        The exponential ansatz is y(x) = a + sign * exp(z(x)) where z(x) is a
        polynomial and "sign" is either +1 or -1 corresponding to decreasing
        and increasing exponentials, respectively. The parameter "sign" is
        automatically deduced from the data.

        It is also assumed that z(x-->inf) = -inf, such that y(x-->inf) --> a.

        If asymptote is None, the ansatz y(x) is fitted with a non-linear
        optimization.

        If asymptote is given and avoid_log=False, a linear fit with respect to
        z(x) := log[sign * (y(x) - asymptote)] is performed.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            asymptote: The infinite-noise limit y(x->inf) (optional argument).
            order: The degree of the polynomial z(x).
            avoid_log: If set to True, the exponential model is not linearized
                with a logarithm and a non-linear fit is applied even if
                asymptote is not None. The default value is False.
            eps: Epsilon to regularize log(sign(scale_factors - asymptote))
                when the argument is to close to zero or negative.
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional information about the
                extrapolated limit is returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ValueError: If the arguments are not consistent with the
                extrapolation model.
            ExtrapolationError: If the extrapolation fit fails.
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """

        # Shift is 0 if asymptote is given, 1 if asymptote is not given
        shift = int(asymptote is None)

        # Check arguments
        error_str = (
            "Data is not enough: at least two data points are necessary."
        )
        if scale_factors is None or exp_values is None:
            raise ValueError(error_str)
        if len(scale_factors) != len(exp_values) or len(scale_factors) < 2:
            raise ValueError(error_str)
        if order > len(scale_factors) - (1 + shift):
            raise ValueError(
                "Extrapolation order is too high. "
                "The order cannot exceed the number"
                f" of data points minus {1 + shift}."
            )
        if not np.allclose(np.real(exp_values), exp_values):
            raise ValueError(
                f"Cannot extrapolate: Some expectation values in {exp_values} "
                f"have non-zero imaginary part."
            )
        exp_values = np.real(exp_values).tolist()

        # Initialize default errors
        zne_error = None
        params_cov = None

        # Deduce "sign" parameter of the exponential ansatz
        linear_params, _ = mitiq_polyfit(scale_factors, exp_values, deg=1)

        if asymptote is not None:
            sign = np.sign(-(asymptote - linear_params[1]))
        else:
            sign = np.sign(-linear_params[0])

        def _ansatz_unknown(x: float, *coeffs: float) -> float:
            """Ansatz of generic order with unknown asymptote."""
            # Coefficients of the polynomial to be exponentiated
            z_coeffs = coeffs[2:][::-1]
            return coeffs[0] + coeffs[1] * np.exp(x * np.polyval(z_coeffs, x))

        def _ansatz_known(x: float, *coeffs: float) -> float:
            """Ansatz of generic order with known asymptote."""
            # Coefficients of the polynomial to be exponentiated
            z_coeffs = coeffs[1:][::-1]
            return asymptote + coeffs[0] * np.exp(x * np.polyval(z_coeffs, x))

        # CASE 1: asymptote is None.
        if asymptote is None:
            # First guess for the parameters
            p_zero = [0.0, sign, -1.0] + [0.0 for _ in range(order - 1)]
            opt_params, params_cov = mitiq_curve_fit(
                _ansatz_unknown, scale_factors, exp_values, p_zero
            )
            # The zero noise limit is ansatz(0)= asympt + b
            zne_limit = opt_params[0] + opt_params[1]

            def zne_curve(scale_factor: float) -> float:
                return _ansatz_unknown(scale_factor, *opt_params)

            # Use propagation of errors to calculate zne_error
            if params_cov is not None:
                if params_cov.shape == (order + 2, order + 2):
                    zne_error = np.sqrt(
                        params_cov[0, 0]
                        + 2 * params_cov[0, 1]
                        + params_cov[1, 1]
                    )

            if full_output:
                return (
                    zne_limit,
                    zne_error,
                    opt_params,
                    params_cov,
                    zne_curve,
                )

            return zne_limit

        # CASE 2: asymptote is given and "avoid_log" is True
        if avoid_log:
            # First guess for the parameters
            p_zero = [sign, -1.0] + [0.0 for _ in range(order - 1)]
            opt_params, params_cov = mitiq_curve_fit(
                _ansatz_known, scale_factors, exp_values, p_zero
            )
            # The zero noise limit is ansatz(0)= asymptote + b
            zne_limit = asymptote + opt_params[0]

            def zne_curve(scale_factor: float) -> float:
                return _ansatz_known(scale_factor, *opt_params)

            # Use propagation of errors to calculate zne_error
            if params_cov is not None:
                if params_cov.shape == (order + 1, order + 1):
                    zne_error = np.sqrt(params_cov[0, 0])

            opt_params = [asymptote] + list(opt_params)

            if full_output:
                return (
                    zne_limit,
                    zne_error,
                    opt_params,
                    params_cov,
                    zne_curve,
                )

            return zne_limit

        # CASE 3: asymptote is given and "avoid_log" is False
        # Polynomial fit of z(x).
        shifted_y = [max(sign * (y - asymptote), eps) for y in exp_values]
        zstack = list(np.log(shifted_y))  # type: ignore

        # Get coefficients {z_j} of z(x)= z_0 + z_1*x + z_2*x**2...
        # Note: coefficients are ordered from high powers to powers of x
        # Weights "w" are used to compensate for error propagation
        # after the log transformation y --> z
        z_coefficients, param_cov = mitiq_polyfit(
            scale_factors,
            zstack,
            deg=order,
            weights=np.sqrt(np.abs(shifted_y)),
        )
        # The zero noise limit is ansatz(0)
        zne_limit = asymptote + sign * np.exp(z_coefficients[-1])

        def _zne_curve(scale_factor: float) -> float:
            return asymptote + sign * np.exp(
                np.polyval(z_coefficients, scale_factor)
            )

        # Use propagation of errors to calculate zne_error
        if params_cov is not None:
            if params_cov.shape == (order + 1, order + 1):
                zne_error = np.exp(z_coefficients[-1]) * np.sqrt(
                    params_cov[order + 1, order + 1]
                )

        # Parameters from low order to high order
        opt_params = [asymptote] + list(z_coefficients[::-1])

        if full_output:
            return zne_limit, zne_error, opt_params, params_cov, _zne_curve
        return zne_limit


# Keep a log of the optimization process storing:
# noise value(s), expectation value(s), parameters, and zero limit
OptimizationHistory = List[
    Tuple[List[Dict[str, float]], List[float], List[float], float]
]


class AdaExpFactory(AdaptiveFactory):
    """Factory object implementing an adaptive zero-noise extrapolation
    algorithm assuming an exponential ansatz y(x) = a + b * exp(-c * x),
    with c > 0.

    The noise scale factors are are chosen adaptively at each step,
    depending on the history of collected results.

    If y(x->inf) is unknown, the ansatz y(x) is fitted with a non-linear
    optimization.

    If y(x->inf) is given and avoid_log=False, the exponential
    model is mapped into a linear model by logarithmic transformation.

    Args:
        steps: The number of optimization steps. At least 3 are necessary.
        scale_factor: The second noise scale factor (the first is always 1.0).
            Further scale factors are adaptively determined.
        asymptote: The infinite-noise limit y(x->inf) (optional argument).
        avoid_log: If set to True, the exponential model is not linearized
            with a logarithm and a non-linear fit is applied even if asymptote
            is not None. The default value is False.
        max_scale_factor: Maximum noise scale factor. Default is 6.0.

    Raises:
        ValueError: If data is not consistent with the extrapolation model.
        ExtrapolationError: If the extrapolation fit fails.
        ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

    """

    _SHIFT_FACTOR = 1.27846
    _EPSILON = 1.0e-9

    def __init__(
        self,
        steps: int,
        scale_factor: float = 2.0,
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        max_scale_factor: float = 6.0,
    ) -> None:
        """Instantiate a new object of this Factory class."""
        super(AdaExpFactory, self).__init__()
        if not (asymptote is None or isinstance(asymptote, float)):
            raise ValueError(
                "The argument 'asymptote' must be either a float or None"
            )
        if scale_factor <= 1:
            raise ValueError(
                "The argument 'scale_factor' must be strictly larger than one."
            )
        if steps < 3 + int(asymptote is None):
            raise ValueError(
                "The argument 'steps' must be an integer"
                " greater or equal to 3. "
                "If 'asymptote' is None, 'steps' must be"
                " greater or equal to 4."
            )
        if max_scale_factor <= 1:
            raise ValueError(
                "The argument 'max_scale_factor' must be"
                " strictly larger than one."
            )
        self._steps = steps
        self._scale_factor = scale_factor
        self.asymptote = asymptote
        self.avoid_log = avoid_log
        self.max_scale_factor = max_scale_factor
        self.history: OptimizationHistory = []

    def next(self) -> Dict[str, float]:
        """Returns a dictionary of parameters to execute a circuit at."""
        # The 1st scale factor is always 1
        if len(self._instack) == 0:
            return {"scale_factor": 1.0}
        # The 2nd scale factor is self._scale_factor
        if len(self._instack) == 1:
            return {"scale_factor": self._scale_factor}
        # If asymptote is None we use 2 * scale_factor as third noise parameter
        if (len(self._instack) == 2) and (self.asymptote is None):
            return {"scale_factor": 2 * self._scale_factor}

        with warnings.catch_warnings():
            # This is an intermediate fit, so we suppress its warning messages
            warnings.simplefilter("ignore", ExtrapolationWarning)
            # Call reduce() to fit the exponent and save it in self.history
            self.reduce()
            # The next line avoids warnings after intermediate extrapolations
            self._already_reduced = False

        # Get the most recent fitted parameters from self.history
        _, _, params, _ = self.history[-1]
        # The exponent parameter is the 3rd element of params
        exponent = params[2]
        # Further noise scale factors are determined with
        # an adaptive rule which depends on self.exponent
        next_scale_factor = min(
            1.0 + self._SHIFT_FACTOR / np.abs(exponent + self._EPSILON),
            self.max_scale_factor,
        )
        return {"scale_factor": next_scale_factor}

    def is_converged(self) -> bool:
        """Returns True if all the needed expectation values have been
        computed, else False.
        """
        if len(self._outstack) != len(self._instack):
            raise IndexError(
                f"The length of 'self._instack' ({len(self._instack)}) "
                f"and 'self._outstack' ({len(self._outstack)}) must be equal."
            )
        return len(self._outstack) == self._steps

    @staticmethod
    def extrapolate(
        scale_factors: Sequence[float],
        exp_values: Sequence[float],
        asymptote: Optional[float] = None,
        avoid_log: bool = False,
        eps: float = 1.0e-6,
        full_output: bool = False,
    ) -> ExtrapolationResult:
        """Static method which evaluates the extrapolation to the zero-noise
        limit assuming an exponential ansatz y(x) = a + b * exp(-c * x),
        with c > 0.

        If y(x->inf) is unknown, the ansatz y(x) is fitted with a non-linear
        optimization.

        If y(x->inf) is given and avoid_log=False, the exponential
        model is mapped into a linear model by a logarithmic transformation.

        Args:
            scale_factors: The array of noise scale factors.
            exp_values: The array of expectation values.
            asymptote: The infinite-noise limit y(x->inf) (optional argument).
            avoid_log: If set to True, the exponential model is not linearized
                with a logarithm and a non-linear fit is applied even if
                asymptote is not None. The default value is False.
            eps: Epsilon to regularize log(sign(scale_factors - asymptote))
                when the argument is to close to zero or negative.
            full_output: If False (default), only the zero-noise limit is
                returned. If True, additional results are returned too.

        Returns:
            The extrapolated zero-noise limit. If full_output is True, also
            returns
            * standard deviation of the extrapolated zero-noise limit,
            * optimal parameters of the best-fit model,
            * parameter covariance matrix of best-fit model,
            * best-fit model as a Callable[[float], float] function.

        Raises:
            ValueError: If the arguments are not consistent with the
                extrapolation model.
            ExtrapolationError: If the extrapolation fit fails.
            ExtrapolationWarning: If the extrapolation fit is ill-conditioned.

        Note:
            This static method computes the zero-noise limit from input
            parameters. To compute the zero-noise limit from the Factory
            parameters, use the ``reduce`` method.
        """
        return ExpFactory.extrapolate(
            scale_factors,
            exp_values,
            asymptote=asymptote,
            avoid_log=avoid_log,
            eps=eps,
            full_output=full_output,
        )

    def reduce(self) -> float:
        """Returns the zero-noise limit found by fitting an exponential
        model to the internal data stored in the factory.

        Returns:
            The zero-noise limit.
        """
        (
            self._zne_limit,
            self._zne_error,
            self._opt_params,
            self._params_cov,
            self._zne_curve,
        ) = self.extrapolate(  # type: ignore [misc]
            self.get_scale_factors(),
            self.get_expectation_values(),
            asymptote=self.asymptote,
            avoid_log=self.avoid_log,
            full_output=True,
        )
        # Update optimization history
        self.history.append(
            (self._instack, self._outstack, self._opt_params, self._zne_limit)
        )
        self._already_reduced = True
        return self._zne_limit
